@@ -3,10 +3,10 @@
  * Mark Nagel <nagel@ics.uci.edu>
  * 20 July 1989
  *
- * $Revision: 7.7 $
+ * $Revision: 7.13 $
  *
  * bool
- * format(fmt, num, buf, buflen)
+ * format(fmt, precision, num, buf, buflen)
  *  char *fmt;
  *  double num;
  *  char buf[];
@@ -54,6 +54,11 @@
  *      in the integer part of the number with numbering beginning
  *      from the right end of the integer.
  *
+ *  &   Precision.  When this character is present in the fractional
+ *      part of the number, it is equavalent to a number of 0's equal
+ *      to the precision specified in the column format command.  For
+ *      example, if the precision is 3, "&" is equivalent to "000".
+ *
  *  \	Quote.  This character causes the next character to be
  *      inserted into the formatted string directly with no
  *      special interpretation.
@@ -97,15 +102,17 @@
 #define MAXBUF	256
 
 static char *fmt_int(char *val, char *fmt, bool comma, bool negative);
-static char *fmt_frac(char *val, char *fmt);
+static char *fmt_frac(char *val, char *fmt, int lprecision);
 static char *fmt_exp(int val, char *fmt);
 
 static void reverse(register char *buf);
 
+char *colformat[COLFORMATS];
+
 /*****************************************************************************/
 
 bool
-format(char *fmt, double val, char *buf, int buflen)
+format(char *fmt, int lprecision, double val, char *buf, int buflen)
 {
     register char *cp;
     char *tmp, *tp;
@@ -235,6 +242,11 @@ format(char *fmt, double val, char *buf, int buflen)
 		case '0':
 		    zero_pad = ++width;
 		    break;
+
+		case '&':
+		    width += lprecision;
+		    zero_pad = width;
+		    break;
 	    }
 	}
 	zero_pad = strlen(decimal) - zero_pad;
@@ -275,7 +287,7 @@ format(char *fmt, double val, char *buf, int buflen)
     }
     ci = strcpy(citmp, ci);
 
-    cf = (fraction) ? fmt_frac(fraction, decimal) : "";
+    cf = (fraction) ? fmt_frac(fraction, decimal, lprecision) : "";
     len_cf = strlen(cf);
     if (len_cf >= cflen) {
     	cflen = len_cf + 40;
@@ -316,10 +328,10 @@ fmt_int(char *val,	/* integer part of the value to be formatted */
  * locate the leftmost digit placeholder
  */
     for (cp = fmt; *cp != EOS; cp++) {
-    if (*cp == '\\')
-	cp++;
-    else if (*cp == '#' || *cp == '0')
-	break;
+	if (*cp == '\\')
+	    cp++;
+	else if (*cp == '#' || *cp == '0')
+	    break;
     }
     digit = (*cp == EOS) ? -1 : cp - fmt;
 
@@ -359,7 +371,8 @@ fmt_int(char *val,	/* integer part of the value to be formatted */
 
 static char *
 fmt_frac(char *val,	/* fractional part of the value to be formatted */
-    char *fmt)		/* fractional portion of format */
+    char *fmt,		/* fractional portion of format */
+    int lprecision)	/* precision, for interpreting the "&" */
 {
     static char buf[MAXBUF];
     register char *bufptr = buf;
@@ -367,7 +380,11 @@ fmt_frac(char *val,	/* fractional part of the value to be formatted */
 
     *bufptr++ = '.';
     while (*fmtptr != EOS) {
-	if (*fmtptr == '\\')
+	if (*fmtptr == '&') {
+	    int i;
+	    for (i = 0; i < lprecision; i++)
+		*bufptr++ = (*valptr != EOS) ? *valptr++ : *fmtptr;
+	} else if (*fmtptr == '\\')
 	    *bufptr++ = *++fmtptr;
 	else if (*fmtptr == '#' || *fmtptr == '0') {
 	    if (*valptr != EOS || *fmtptr == '0')
@@ -471,84 +488,78 @@ bool
 engformat(int fmt, int width, int lprecision, double val, char *buf, int buflen)
 {
 
-  static char *engmult[] = {
-      "-18", "-15", "-12", "-09", "-06", "-03",
-      "+00",
-      "+03", "+06", "+09", "+12", "+15", "+18"
-  };
-  int engind = 0;
-  double engmant, pow(), engabs, engexp;
-  if (buflen < width) return (false);
-  if (fmt == REFMTFIX)
-  (void) sprintf(buf,"%*.*f", width, lprecision, val);
-  if (fmt == REFMTFLT)
-  (void) sprintf(buf,"%*.*E", width, lprecision, val);
-  if (fmt == REFMTENG)
-  {
-    if (val == 0e0)    /* Hack to get zeroes to line up in engr fmt */
-    {
-      (void) sprintf((buf-1),"%*.*f ", width, lprecision, val);
-    }
-    else
-    {
-      engabs = (val);
-      if (engabs < 0e0) engabs = -engabs;
-      if ((engabs >= 1e-18) && (engabs < 1e-15)) engind=0;
-      if ((engabs >= 1e-15) && (engabs < 1e-12)) engind=1;
-      if ((engabs >= 1e-12) && (engabs < 1e-9)) engind=2;
-      if ((engabs >= 1e-9) && (engabs < 1e-6)) engind=3;
-      if ((engabs >= 1e-6) && (engabs < 1e-3)) engind=4;
-      if ((engabs >= 1e-3) && (engabs < 1)) engind=5;
-      if ((engabs >= 1) && (engabs < 1e3)) engind=6;
-      if ((engabs >= 1e3) && (engabs < 1e6)) engind=7;
-      if ((engabs >= 1e6) && (engabs < 1e9)) engind=8;
-      if ((engabs >= 1e9) && (engabs < 1e12)) engind=9;
-      if ((engabs >= 1e12) && (engabs < 1e15)) engind=10;
-      if ((engabs >= 1e15) && (engabs < 1e18)) engind=11;
-      if ((engabs >= 1e18) && (engabs < 1e21)) engind=12;
-      if ((engabs < 1e-18) || (engabs >= 1e21))
-      {
-      /* Revert to floating point */
-        (void) sprintf(buf,"%*.*E", width, lprecision, val);
-      }
-      else
-      {
-        engexp = (double) (engind-6)*3;
-        engmant = val/pow(10.0e0,engexp);
-        (void) sprintf(buf,"%*.*fe%s", width-4,
-                      lprecision, engmant, engmult[engind]);
-      }
-    }
-  }
-  if (fmt == REFMTDATE) {
-    int i;
-    time_t secs;
+    static char *engmult[] = {
+	"-18", "-15", "-12", "-09", "-06", "-03",
+	"+00",
+	"+03", "+06", "+09", "+12", "+15", "+18"
+    };
+    int engind = 0;
+    double engmant, pow(), engabs, engexp;
 
-    if (buflen < 9) {
-      for (i = 0; i < width; i++) buf[i] = '*';
-      buf[i] = '\0';
+    if (buflen < width) return (false);
+    if (fmt >= 0 && fmt < COLFORMATS && colformat[fmt])
+	return (format(colformat[fmt], lprecision, val, buf, buflen));
+    if (fmt == REFMTFIX)
+	(void) sprintf(buf,"%*.*f", width, lprecision, val);
+    if (fmt == REFMTFLT)
+	(void) sprintf(buf,"%*.*E", width, lprecision, val);
+    if (fmt == REFMTENG) {
+	if (val == 0e0) {	/* Hack to get zeroes to line up in engr fmt */
+	    (void) sprintf((buf-1),"%*.*f ", width, lprecision, val);
+	} else {
+	    engabs = (val);
+	    if ( engabs <  0e0)       engabs = -engabs;
+	    if ((engabs >= 1e-18) && (engabs <  1e-15)) engind=0;
+	    if ((engabs >= 1e-15) && (engabs <  1e-12)) engind=1;
+	    if ((engabs >= 1e-12) && (engabs <  1e-9 )) engind=2;
+	    if ((engabs >= 1e-9)  && (engabs <  1e-6 )) engind=3;
+	    if ((engabs >= 1e-6)  && (engabs <  1e-3 )) engind=4;
+	    if ((engabs >= 1e-3)  && (engabs <  1    )) engind=5;
+	    if ((engabs >= 1)     && (engabs <  1e3  )) engind=6;
+	    if ((engabs >= 1e3)   && (engabs <  1e6  )) engind=7;
+	    if ((engabs >= 1e6)   && (engabs <  1e9  )) engind=8;
+	    if ((engabs >= 1e9)   && (engabs <  1e12 )) engind=9;
+	    if ((engabs >= 1e12)  && (engabs <  1e15 )) engind=10;
+	    if ((engabs >= 1e15)  && (engabs <  1e18 )) engind=11;
+	    if ((engabs >= 1e18)  && (engabs <  1e21 )) engind=12;
+	    if ((engabs < 1e-18)  || (engabs >= 1e21 )) {
+		/* Revert to floating point */
+		(void) sprintf(buf,"%*.*E", width, lprecision, val);
+	    } else {
+		engexp = (double) (engind-6)*3;
+		engmant = val/pow(10.0e0,engexp);
+		(void) sprintf(buf,"%*.*fe%s", width-4,
+			lprecision, engmant, engmult[engind]);
+	    }
+	}
     }
-    else {
-      secs = (time_t)val;
-      strftime(buf,buflen,"%e %b %y",localtime(&secs));
-      for (i = 9; i < width; i++) buf[i] = ' ';
-      buf[i] = '\0';
-    }
-  }
-  if (fmt == REFMTLDATE) {
-    int i;
-    time_t secs;
+    if (fmt == REFMTDATE) {
+	int i;
+	time_t secs;
 
-    if (buflen < 11) {
-      for (i = 0; i < width; i++) buf[i] = '*';
-      buf[i] = '\0';
+	if (buflen < 9) {
+	    for (i = 0; i < width; i++) buf[i] = '*';
+	    buf[i] = '\0';
+	} else {
+	    secs = (time_t)val;
+	    strftime(buf,buflen,"%e %b %y",localtime(&secs));
+	    for (i = 9; i < width; i++) buf[i] = ' ';
+	    buf[i] = '\0';
+	}
     }
-    else {
-      secs = (time_t)val;
-      strftime(buf,buflen,"%e %b %Y",localtime(&secs));
-      for (i = 11; i < width; i++) buf[i] = ' ';
-      buf[i] = '\0';
+    if (fmt == REFMTLDATE) {
+	int i;
+	time_t secs;
+
+	if (buflen < 11) {
+	    for (i = 0; i < width; i++) buf[i] = '*';
+	    buf[i] = '\0';
+	} else {
+	    secs = (time_t)val;
+	    strftime(buf,buflen,"%e %b %Y",localtime(&secs));
+	    for (i = 11; i < width; i++) buf[i] = ' ';
+	    buf[i] = '\0';
+	}
     }
-  }
-  return (true);
+    return (true);
 }
