@@ -2,13 +2,13 @@
  * Help functions for sc 
  * R. Bond, 1988
  * J. Buhrt 1990
- * $Revision: 7.7 $
+ * $Revision: 7.13 $
  */
 
 #ifdef QREF
 #include <stdio.h>
 char	*header = " Quick Reference";
-char	*revision = "$Revision: 7.7 $";
+char	*revision = "$Revision: 7.13 $";
 #else
 #include <curses.h>
 #include "sc.h"
@@ -254,7 +254,7 @@ char *row[] = {
 "                 Insert rows, columns or merge the cells.",
 "     vr, vc      Remove expressions from the affected rows (columns),",
 "                 leaving only the values.",
-"     zr, zc      Hide (``zap'') the current row (column)",
+"     Zr, Zc      Hide (``zap'') the current row (column)",
 "     sr, sc      Show hidden rows (columns)",
 "     f           Set the output format to be used with the values of",
 "                 each cell in this column.  Enter field width and",
