@@ -7,7 +7,7 @@
  *
  *              More mods Robert Bond, 12/86
  *		More mods by Alan Silverstein, 3-4/88, see list of changes.
- *		$Revision: 7.7 $
+ *		$Revision: 7.13 $
  */
 
 #include <sys/types.h>
@@ -41,6 +41,10 @@
 #include <curses.h>
 #include "sc.h"
 
+#ifndef MSDOS
+#include <unistd.h>
+#endif
+
 #if defined(RE_COMP)
 extern char *re_comp(char *s);
 extern char *re_exec(char *s);
@@ -60,10 +64,12 @@ char *regex();
 struct go_save gs;
 
 /* g_type can be: */
-#define G_NONE 0			/* Starting value - must be 0 */
+#define G_NONE 0	/* Starting value - must be 0 */
 #define G_NUM 1
 #define G_STR 2
-#define G_CELL 3
+#define G_NSTR 3
+#define G_XSTR 4
+#define G_CELL 5
 
 #define ISVALID(r,c)	((r)>=0 && (r)<maxrows && (c)>=0 && (c)<maxcols)
 
@@ -77,10 +83,8 @@ int	gmyrow, gmycol;	/* globals used to implement @myrow, @mycol cmds */
 /* a linked list of free [struct enodes]'s, uses .e.o.left as the pointer */
 struct enode *freeenodes = NULL;
 
-char	*seval(register struct enode *se);
 double	dolookup(struct enode * val, int minr, int minc, int maxr,
 	int maxc, int offr, int offc);
-double	eval(register struct enode *e);
 double	fn1_eval(double (*fn)(), double arg);
 double	fn2_eval(double (*fn)(), double arg1, double arg2);
 int	RealEvalAll();
@@ -553,7 +557,7 @@ doeqs(char *s1, char *s2)
 
 
 /*
- * Given a string representing a column name and a value which is a column
+ * Given a string representing a column name and a value which is a row
  * number, return a pointer to the selected cell's entry, if any, else NULL.
  * Use only the integer part of the column number.  Always free the string.
  */
@@ -664,13 +668,13 @@ eval(register struct enode *e)
 			denom = eval(e->e.o.right);
 			if (cellerror) {
 			    cellerror = CELLINVALID;
-			    return((double) 0);
+			    return ((double) 0);
 			} else
 			if (denom)
-			    return(num/denom);
+			    return (num/denom);
 			else {
 			    cellerror = CELLERROR;
-			    return((double) 0);
+			    return ((double) 0);
 			}
 	}
 	case '%':	{
@@ -708,8 +712,14 @@ eval(register struct enode *e)
 	case 'm':	return (-eval(e->e.o.right));
 	case 'f':	return (eval(e->e.o.right));
 	case '~':	return (eval(e->e.o.right) == 0.0);
+	case ';':	return (((int)eval(e->e.o.left) & 7) +
+				(((int)eval(e->e.o.right) & 7) << 3));
 	case O_CONST:	return (e->e.k);
-	case O_VAR:	if (e->e.v.vp->cellerror)
+	case O_VAR:	if (!e->e.v.vp || e->e.v.vp->flags & is_deleted) {
+			    cellerror = CELLERROR;
+			    return (double) 0;
+			}
+			if (e->e.v.vp->cellerror)
 			    cellerror = CELLINVALID;
 			return (e->e.v.vp->v);
 	case INDEX:
@@ -744,6 +754,8 @@ eval(register struct enode *e)
  	case REDUCE | 'a':
  	case REDUCE | 'c':
  	case REDUCE | 's':
+ 	case REDUCE | 'R':
+ 	case REDUCE | 'C':
 	case REDUCE | MAX:
 	case REDUCE | MIN:
 	    {	int r, c;
@@ -761,6 +773,8 @@ eval(register struct enode *e)
  	            case REDUCE | 'a': return doavg(minr, minc, maxr, maxc);
  	            case REDUCE | 'c': return docount(minr, minc, maxr, maxc);
  	            case REDUCE | 's': return dostddev(minr, minc, maxr, maxc);
+ 	            case REDUCE | 'R': return (maxr - minr + 1);
+ 	            case REDUCE | 'C': return (maxc - minc + 1);
  	            case REDUCE | MAX: return domax(minr, minc, maxr, maxc);
  	            case REDUCE | MIN: return domin(minr, minc, maxr, maxc);
 		}
@@ -789,7 +803,7 @@ eval(register struct enode *e)
 		return rint(eval(e->e.o.right));
 	    else {
 		double temp = eval(e->e.o.right);
-		return(temp-floor(temp) < 0.5 ? floor(temp) : ceil(temp));
+		return (temp - floor(temp) < 0.5 ? floor(temp) : ceil(temp));
 	    }
  	case ROUND:
 	    {	int prec = (int) eval(e->e.o.right);
@@ -804,16 +818,16 @@ eval(register struct enode *e)
 		else {
 		    double temp = eval(e->e.o.left);
 		    temp *= scal;
-		    temp = ((temp-floor(temp)) < 0.5 ?
+		    temp = ((temp - floor(temp)) < 0.5 ?
 			    floor(temp) : ceil(temp));
-		    return(temp / scal);
+		    return (temp / scal);
 		}
 	    }
 	case FV:
 	case PV:
-	case PMT:	return(finfunc(e->op,eval(e->e.o.left),
-				   eval(e->e.o.right->e.o.left),
-				      eval(e->e.o.right->e.o.right)));
+	case PMT:	return (finfunc(e->op,eval(e->e.o.left),
+				    eval(e->e.o.right->e.o.left),
+				    eval(e->e.o.right->e.o.right)));
 	case HOUR:	return (dotime(HOUR, eval(e->e.o.right)));
 	case MINUTE:	return (dotime(MINUTE, eval(e->e.o.right)));
 	case SECOND:	return (dotime(SECOND, eval(e->e.o.right)));
@@ -822,11 +836,11 @@ eval(register struct enode *e)
 	case YEAR:	return (dotime(YEAR, eval(e->e.o.right)));
 	case NOW:	return (dotime(NOW, (double)0.0));
 	case DTS:	return (dodts((int)eval(e->e.o.left),
-				 (int)eval(e->e.o.right->e.o.left),
-				 (int)eval(e->e.o.right->e.o.right)));
+				    (int)eval(e->e.o.right->e.o.left),
+				    (int)eval(e->e.o.right->e.o.right)));
 	case TTS:	return (dotts((int)eval(e->e.o.left),
-				 (int)eval(e->e.o.right->e.o.left),
-				 (int)eval(e->e.o.right->e.o.right)));
+				    (int)eval(e->e.o.right->e.o.left),
+				    (int)eval(e->e.o.right->e.o.right)));
 	case STON:	return (doston(seval(e->e.o.right)));
 	case EQS:       return (doeqs(seval(e->e.o.right),seval(e->e.o.left)));
 	case LMAX:	return dolmax(e);
@@ -835,11 +849,22 @@ eval(register struct enode *e)
 	case MYROW:	return ((double) gmyrow);
 	case MYCOL:	return ((double) gmycol);
 	case NUMITER:	return ((double) repct);
+	case ERR_:	cellerror = CELLERROR;
+			return ((double) 0);
+	case PI_:	return ((double) PI);
+	case BLACK:	return ((double) COLOR_BLACK);
+	case RED:	return ((double) COLOR_RED);
+	case GREEN:	return ((double) COLOR_GREEN);
+	case YELLOW:	return ((double) COLOR_YELLOW);
+	case BLUE:	return ((double) COLOR_BLUE);
+	case MAGENTA:	return ((double) COLOR_MAGENTA);
+	case CYAN:	return ((double) COLOR_CYAN);
+	case WHITE:	return ((double) COLOR_WHITE);
 	default:	error ("Illegal numeric expression");
 			exprerr = 1;
     }
     cellerror = CELLERROR;
-    return((double)0.0);
+    return ((double)0.0);
 }
 
 #ifdef SIGVOID
@@ -865,7 +890,8 @@ eval_fpe() /* Trap for FPE errors in eval */
     longjmp(fpe_save, 1);
 }
 
-double fn1_eval(double (*fn)(), double arg)
+double
+fn1_eval(double (*fn)(), double arg)
 {
     double res;
     errno = 0;
@@ -876,7 +902,8 @@ double fn1_eval(double (*fn)(), double arg)
     return res;
 }
 
-double fn2_eval(double (*fn)(), double arg1, double arg2)
+double
+fn2_eval(double (*fn)(), double arg1, double arg2)
 {
     double res;
     errno = 0;
@@ -974,30 +1001,29 @@ doext(char *command, double value)
     char buff[FBUFLEN];		/* command line/return, not permanently alloc */
 
     if (!extfunc) {
-	error ("Warning: external functions disabled; using %s value",
+	error("Warning: external functions disabled; using %s value",
 		((prevstr == NULL) || (*prevstr == '\0')) ?
 			"null" : "previous");
 
-	if (command) scxfree (command);
+	if (command) scxfree(command);
     } else {
 	if ((! command) || (! *command)) {
 	    error ("Warning: external function given null command name");
 	    cellerror = CELLERROR;
-	    if (command) scxfree (command);
+	    if (command) scxfree(command);
 	} else {
 	    FILE *pp;
 
-	    (void) sprintf (buff, "%s %g", command, value); /* build cmd line */
-	    scxfree (command);
+	    (void) sprintf(buff, "%s %g", command, value); /* build cmd line */
+	    scxfree(command);
 
-	    error ("Running external function...");
+	    error("Running external function...");
 	    (void) refresh();
 
 	    if ((pp = popen(buff, "r")) == (FILE *) NULL) {	/* run it */
-		error ("Warning: running \"%s\" failed", buff);
+		error("Warning: running \"%s\" failed", buff);
 		cellerror = CELLERROR;
-	    }
-	    else {
+	    } else {
 		if (fgets(buff, sizeof(buff)-1, pp) == NULL)	/* one line */
 		    error("Warning: external function returned nothing");
 		else {
@@ -1009,20 +1035,20 @@ doext(char *command, double value)
 		    if ((cp = strchr(buff, '\n')))	/* contains newline */
 			*cp = '\0';			/* end string there */
 
-		    if (strlen(buff) + 1 > prevlen)
-		    {	prevlen = strlen(buff) + 40;
+		    if (strlen(buff) + 1 > prevlen) {
+		    	prevlen = strlen(buff) + 40;
 			prevstr = scxrealloc(prevstr, prevlen);
 		    }
 		    (void) strcpy (prevstr, buff);
 			 /* save alloc'd copy */
 		}
-		(void) pclose (pp);
+		(void) pclose(pp);
 
 	    } /* else */
 	} /* else */
     } /* else */
     if (prevstr)
-	return (strcpy (scxmalloc ((unsigned) (strlen (prevstr) + 1)), prevstr));
+	return (strcpy(scxmalloc((unsigned) (strlen(prevstr) + 1)), prevstr));
     else
 	return (strcpy(scxmalloc((unsigned)1), ""));
 }
@@ -1191,7 +1217,7 @@ seval(register struct enode *se)
 			    (int)eval(se->e.o.right->e.o.left) - 1,
 			    (int)eval(se->e.o.right->e.o.right) - 1));
 	case COLTOA: return (strcpy(scxmalloc((unsigned)10),
-				   coltoa((int)eval(se->e.o.right)+1)));
+				   coltoa((int)eval(se->e.o.right))));
 	default:
 		     error("Illegal string expression");
 		     exprerr = 1;
@@ -1226,14 +1252,36 @@ setiterations(int i)
 void
 EvalAll()
 {
-    int lastcnt;
+    int lastcnt, pair, v;
   
     repct = 1;
     (void) signal(SIGFPE, eval_fpe);
 
     while ((lastcnt = RealEvalAll()) && (++repct <= propagation));
     if ((propagation > 1) && (lastcnt > 0))
- 	error("Still changing after %d iterations",repct-1);
+ 	error("Still changing after %d iterations", repct - 1);
+
+    if (usecurses && color && has_colors()) {
+	for (pair = 0; pair < 8; pair++) {
+	    cellerror = CELLOK;
+	    if (cpairs[pair] && cpairs[pair]->expr) {
+		v = (int) eval(cpairs[pair]->expr);
+		cpairs[pair]->fg = v & 7;
+		cpairs[pair]->bg = (v >> 3) & 7;
+		init_pair(pair + 1, cpairs[pair]->fg, cpairs[pair]->bg);
+	    }
+	    /* Can't see to fix the problem if color 1 has an error, so
+	     * turn off color in that case.
+	     */
+	    if (pair == 0 && cellerror) {
+		color = 0;
+		attron(COLOR_PAIR(0));
+		color_set(0, NULL);
+		error("Error in color 1: color turned off");
+	    }
+	}
+
+    }
 
     (void) signal(SIGFPE, doquit);
 }
@@ -1253,12 +1301,14 @@ RealEvalAll()
     if (calc_order == BYROWS) {
 	for (i=0; i<=maxrow; i++)
 	    for (j=0; j<=maxcol; j++)
-		if ((p = *ATBL(tbl,i,j)) && !(p->flags&is_locked) && p->expr) RealEvalOne(p,i,j,&chgct);
+		if ((p = *ATBL(tbl,i,j)) && p->expr)
+		    RealEvalOne(p,i,j,&chgct);
     }
     else if (calc_order == BYCOLS) {
-	for (j=0; j<=maxcol; j++)
-	{   for (i=0; i<=maxrow; i++)
-		if ((p = *ATBL(tbl,i,j)) && !(p->flags&is_locked) && p->expr) RealEvalOne(p,i,j,&chgct);
+	for (j=0; j<=maxcol; j++) {
+	    for (i=0; i<=maxrow; i++)
+		if ((p = *ATBL(tbl,i,j)) && p->expr)
+		    RealEvalOne(p,i,j,&chgct);
 	}
     }
     else error("Internal error calc_order");
@@ -1269,6 +1319,8 @@ RealEvalAll()
 void
 RealEvalOne(register struct ent *p, int i, int j, int *chgct)
 {
+    gmyrow=i; gmycol=j;
+
     if (p->flags & is_strexpr) {
 	char *v;
 	if (setjmp(fpe_save)) {
@@ -1287,7 +1339,7 @@ RealEvalOne(register struct ent *p, int i, int j, int *chgct)
 	    p->flags |= is_changed;
 	    changed++;
 	}
-	if(p->label)
+	if (p->label)
 	    scxfree(p->label);
 	p->label = v;
     } else {
@@ -1298,7 +1350,6 @@ RealEvalOne(register struct ent *p, int i, int j, int *chgct)
 	    v = (double)0.0;
 	} else {
 	    cellerror = CELLOK;
-	    gmyrow=i; gmycol=j;
 	    v = eval(p->expr);
 	}
 	/* The following line is necessary to force gcc to clear the
@@ -1306,7 +1357,7 @@ RealEvalOne(register struct ent *p, int i, int j, int *chgct)
 	 * they really are due to comparing a double in memory with a
 	 * higher precision value in the FPU, causing the dreaded
 	 * "Still changing after x iterations" message.  It may be
-	 * necessary for other compilers as well.
+	 * necessary for other compilers as well. - CRM
 	 */
 	fflush(stdout);
 	if ((cellerror != p->cellerror) || (v != p->v)) {
@@ -1368,11 +1419,10 @@ struct enode *
 new_const(int op, double a1)
 {
     register struct enode *p;
-    if (freeenodes)	/* reuse an already free'd enode */
-    {	p = freeenodes;
+    if (freeenodes) {	/* reuse an already free'd enode */
+	p = freeenodes;
 	freeenodes = p->e.o.left;
-    }
-    else
+    } else
 	p = (struct enode *) scxmalloc ((unsigned)sizeof(struct enode));
     p->op = op;
     p->e.k = a1;
@@ -1384,11 +1434,10 @@ new_str(char *s)
 {
     register struct enode *p;
 
-    if (freeenodes)	/* reuse an already free'd enode */
-    {	p = freeenodes;
+    if (freeenodes) {	/* reuse an already free'd enode */
+    	p = freeenodes;
 	freeenodes = p->e.o.left;
-    }
-    else
+    } else
 	p = (struct enode *) scxmalloc ((unsigned)sizeof(struct enode));
     p->op = O_SCONST;
     p->e.s = s;
@@ -1398,8 +1447,8 @@ new_str(char *s)
 void
 copy(struct ent *dv1, struct ent *dv2, struct ent *v1, struct ent *v2)
 {
-    int minsr, minsc;
-    int maxsr, maxsc;
+    static int minsr = -1, minsc = -1;
+    static int maxsr = -1, maxsc = -1;
     int mindr, mindc;
     int maxdr, maxdc;
     int vr, vc;
@@ -1411,14 +1460,20 @@ copy(struct ent *dv1, struct ent *dv2, struct ent *v1, struct ent *v2)
     maxdc = dv2->col;
     if (mindr>maxdr) r = maxdr, maxdr = mindr, mindr = r;
     if (mindc>maxdc) c = maxdc, maxdc = mindc, mindc = c;
-    maxsr = v2->row;
-    maxsc = v2->col;
-    minsr = v1->row;
-    minsc = v1->col;
-    if (minsr>maxsr) r = maxsr, maxsr = minsr, minsr = r;
-    if (minsc>maxsc) c = maxsc, maxsc = minsc, minsc = c;
+
+    if (v1 != NULL) {
+	maxsr = v2->row;
+	maxsc = v2->col;
+	minsr = v1->row;
+	minsc = v1->col;
+	if (minsr>maxsr) r = maxsr, maxsr = minsr, minsr = r;
+	if (minsc>maxsc) c = maxsc, maxsc = minsc, minsc = c;
+    } else
+	if (minsr == -1) return;
+
     checkbounds(&maxdr, &maxdc);
 
+    flush_saved();
     erase_area(mindr, mindc, maxdr, maxdc);
     if (minsr == maxsr && minsc == maxsc) {
 	/* Source is a single cell */
@@ -1454,10 +1509,8 @@ copyrtv(int vr, int vc, int minsr, int minsc, int maxsr, int maxsc)
 	    	n = lookat(dr, dc);
 		if (n->flags&is_locked) continue;
 		(void) clearent(n);
-		copyent( n, p, dr - sr, dc - sc);
-	    }
-	    else
-	    if ((n = *ATBL(tbl, dr, dc)))
+		copyent(n, p, dr - sr, dc - sc);
+	    } else if ((n = *ATBL(tbl, dr, dc)))
 		(void) clearent(n);
 	}
 }
@@ -1466,10 +1519,29 @@ copyrtv(int vr, int vc, int minsr, int minsc, int maxsr, int maxsc)
 void
 eraser(struct ent *v1, struct ent *v2)
 {
-    FullUpdate++;
     flush_saved();
     erase_area(v1->row, v1->col, v2->row, v2->col);
     sync_refs();
+    FullUpdate++;
+    modflg++;
+}
+
+/* YANK a Range of cells */
+void
+yankr(struct ent *v1, struct ent *v2)
+{
+    flush_saved();
+    yank_area(v1->row, v1->col, v2->row, v2->col);
+}
+
+/* MOVE a Range of cells */
+void
+mover(struct ent *d, struct ent *v1, struct ent *v2)
+{
+    flush_saved();
+    move_area(d->row, d->col, v1->row, v1->col, v2->row, v2->col);
+    sync_refs();
+    FullUpdate++;
 }
 
 /* Goto subroutines */
@@ -1478,7 +1550,8 @@ void
 g_free()
 {
     switch (gs.g_type) {
-	case G_STR: scxfree(gs.g_s); break;
+	case G_STR:
+	case G_NSTR: scxfree(gs.g_s); break;
 	default: break;
     }
     gs.g_type = G_NONE;
@@ -1489,11 +1562,7 @@ g_free()
 void
 go_last()
 {
-    /* Remember the current position */
-    savedrow[0] = currow;
-    savedcol[0] = curcol;
-    savedstrow[0] = strow;
-    savedstcol[0] = stcol;
+    int num = 0;
 
     switch (gs.g_type) {
 	case G_NONE:
@@ -1502,14 +1571,18 @@ go_last()
 	    num_search(gs.g_n, gs.g_row, gs.g_col,
 		gs.g_lastrow, gs.g_lastcol, gs.errsearch);
 	    break;
-	case  G_CELL:
+	case G_CELL:
 	    moveto(gs.g_row, gs.g_col, gs.g_lastrow, gs.g_lastcol,
 		gs.strow, gs.stcol);
 	    break;
-	case  G_STR: 
+	case G_XSTR: 
+	    num++;
+	case G_NSTR: 
+	    num++;
+	case G_STR: 
 	    gs.g_type = G_NONE;	/* Don't free the string */
-	    str_search(gs.g_s, gs.g_row, gs.g_col,
-		gs.g_lastrow, gs.g_lastcol); 
+	    str_search(gs.g_s, gs.g_row, gs.g_col, gs.g_lastrow, gs.g_lastcol,
+		    num); 
 	    break;
 
 	default: error("go_last: internal error");
@@ -1526,11 +1599,13 @@ moveto(int row, int col, int lastrow, int lastcol, int cornerrow,
 {
     register int i;
 
-    /* Remember the current position */
-    savedrow[0] = currow;
-    savedcol[0] = curcol;
-    savedstrow[0] = strow;
-    savedstcol[0] = stcol;
+    if (row != -1 && (row != currow || col != curcol)) {
+	/* Remember the current position */
+	savedrow[0] = currow;
+	savedcol[0] = curcol;
+	savedstrow[0] = strow;
+	savedstcol[0] = stcol;
+    }
 
     currow = row;
     curcol = col;
@@ -1559,6 +1634,10 @@ moveto(int row, int col, int lastrow, int lastcol, int cornerrow,
 	colsinrange += fwidth[i];
     }
     FullUpdate++;
+    if (loading) {
+	update(1);
+	changed = 0;
+    }
 }
 
 /*
@@ -1628,16 +1707,20 @@ num_search(double n, int firstrow, int firstcol, int lastrow,
     curcol = c;
     rowsinrange = 1;
     colsinrange = fwidth[curcol];
+    if (loading) {
+	update(1);
+	changed = 0;
+    }
 }
 
 /* 'goto' a cell containing a matching string */
 void
-str_search(char *s, int firstrow, int firstcol, int lastrow,
-    int lastcol)
+str_search(char *s, int firstrow, int firstcol, int lastrow, int lastcol,
+	int num)
 {
     struct ent	*p;
-    int	r, c;
-    int	endr, endc;
+    int		r, c;
+    int		endr, endc;
     char	*tmp;
 
     /* Remember the current position */
@@ -1662,7 +1745,7 @@ str_search(char *s, int firstrow, int firstcol, int lastrow,
     }
 #endif
     g_free();
-    gs.g_type = G_STR;
+    gs.g_type = G_STR + num;
     gs.g_s = s;
     gs.g_row = firstrow;
     gs.g_col = firstcol;
@@ -1691,17 +1774,61 @@ str_search(char *s, int firstrow, int firstcol, int lastrow,
 	    }
 	}
 	p = *ATBL(tbl, r, c);
-	if (!col_hidden[c] && p && (p->label)
+	if (gs.g_type == G_NSTR) {
+	    *line = '\0';
+	    if (p) {
+		if (p->cellerror)
+		    sprintf(line, "%s", p->cellerror == CELLERROR ?
+			    "ERROR" : "INVALID");
+		else if (p->flags & is_valid) {
+		    if (p->format) {
+			if (*(p->format) == ctl('d')) {
+			    time_t i = (time_t) (p->v);
+			    strftime(line, sizeof(line), (p->format)+1,
+				localtime(&i));
+			} else
+			    format(p->format, precision[c], p->v, line,
+				    sizeof(line));
+		    } else
+			engformat(realfmt[c], fwidth[c], precision[c],
+				p->v, line, sizeof(line));
+		}
+	    }
+	} else if (gs.g_type == G_XSTR) {
+	    *line = '\0';
+	    if (p && p->expr) {
+		linelim = 0;
+		decompile(p->expr, 0);	/* set line to expr */
+		line[linelim] = '\0';
+		if (*line == '?')
+		    *line = '\0';
+	    }
+	}
+	if (!col_hidden[c])
+	    if (gs.g_type == G_STR) {
+		if (p && p->label
 #if defined(RE_COMP)
-		  			&& (re_exec(p->label) != 0))
+			&& (re_exec(p->label) != 0))
 #else
 #if defined(REGCMP)
-				&& (regex(tmp, p->label) != (char *)0))
+			&& (regex(tmp, p->label) != (char *)0))
 #else
-					&& (strcmp(s, p->label) == 0))
+			&& (strcmp(s, p->label) == 0))
 #endif
 #endif
-	    break;
+		    break;
+	    } else			/* gs.g_type != G_STR */
+		if (*line != '\0'
+#if defined(RE_COMP)
+			&& (re_exec(line) != 0))
+#else
+#if defined(REGCMP)
+			&& (regex(tmp, line) != (char *)0))
+#else
+			&& (strcmp(s, line) == 0))
+#endif
+#endif
+		    break;
 	if (r == endr && c == endc) {
 	    error("String not found");
 #if defined(REGCMP)
@@ -1717,6 +1844,10 @@ str_search(char *s, int firstrow, int firstcol, int lastrow,
 #if defined(REGCMP)
     free(tmp);
 #endif
+    if (loading) {
+	update(1);
+	changed = 0;
+    }
 }
 
 /* fill a range with constants */
@@ -1825,27 +1956,29 @@ let(struct ent *v, struct enode *e)
     double val;
     unsigned isconstant = constant(e);
 
+    if (locked_cell(v->row, v->col))
+	return;
     if (v->row == currow && v->col == curcol)
 	cellassign = 1;
     if (loading && !isconstant)
 	val = (double)0.0;
-    else
-    {
+    else {
 	exprerr = 0;
 	(void) signal(SIGFPE, eval_fpe);
 	if (setjmp(fpe_save)) {
-	    error ("Floating point exception in cell %s", v_name(v->row, v->col));
+	    error("Floating point exception in cell %s", v_name(v->row, v->col));
 	    val = (double)0.0;
 	    cellerror = CELLERROR;
 	} else {
 	    cellerror = CELLOK;
 	    val = eval(e);
 	}
-	if (v->cellerror != cellerror)
-	{	v->flags |= is_changed;
-		changed++;	modflg++;
-		FullUpdate++;
-		v->cellerror = cellerror;
+	if (v->cellerror != cellerror) {
+	    v->flags |= is_changed;
+	    changed++;
+	    modflg++;
+	    FullUpdate++;
+	    v->cellerror = cellerror;
 	}
 	(void) signal(SIGFPE, doquit);
 	if (exprerr) {
@@ -1872,9 +2005,7 @@ let(struct ent *v, struct enode *e)
 	    v->expr = (struct enode *)0;
 	}
 	efree(e);
-    }
-    else
-    {
+    } else {
 	efree(v->expr);
 	v->expr = e;
 	v->flags &= ~is_strexpr;
@@ -1889,7 +2020,10 @@ slet(struct ent *v, struct enode *se, int flushdir)
 {
     char *p;
 
-    cellassign = 1;
+    if (locked_cell(v->row, v->col))
+	return;
+    if (v->row == currow && v->col == curcol)
+	cellassign = 1;
     exprerr = 0;
     (void) signal(SIGFPE, eval_fpe);
     if (setjmp(fpe_save)) {
@@ -1900,8 +2034,8 @@ slet(struct ent *v, struct enode *se, int flushdir)
 	cellerror = CELLOK;
 	p = seval(se);
     }
-    if (v->cellerror != cellerror)
-    {	v->flags |= is_changed;
+    if (v->cellerror != cellerror) {
+    	v->flags |= is_changed;
 	changed++;	modflg++;
 	FullUpdate++;
 	v->cellerror = cellerror;
@@ -1956,14 +2090,13 @@ format_cell(struct ent *v1, struct ent *v2, char *s)
     if (minc < 0) minc = 0;
 
     FullUpdate++;
-	modflg++;
+    modflg++;
+
     for (r = minr; r <= maxr; r++)
 	for (c = minc; c <= maxc; c++) {
 	    n = lookat(r, c);
-	    if (n->flags&is_locked) {
-		error("Cell %s%d is locked", coltoa(n->col), n->row);
+	    if (locked_cell(n->row, n->col))
 		continue;
-	    }
 	    if (n->format)
 		scxfree(n->format);
 	    n->format = 0;
@@ -1980,11 +2113,10 @@ hide_row(int arg)
 	error("Invalid Range");
 	return;
     }
-    if (arg >= maxrows-1)
-    {
-	if (!growtbl(GROWROW, arg+1, 0))
-	{	error("You can't hide the last row");
-		return;
+    if (arg >= maxrows-1) {
+	if (!growtbl(GROWROW, arg+1, 0)) {
+	    error("You can't hide the last row");
+	    return;
 	}
     }
     FullUpdate++;
@@ -1998,10 +2130,10 @@ hide_col(int arg)
 	error("Invalid Range");
 	return;
     }
-    if (arg >= maxcols-1)
-    {	if ((arg >= ABSMAXCOLS-1) || !growtbl(GROWCOL, 0, arg+1))
-	{	error("You can't hide the last col");
-		return;
+    if (arg >= maxcols-1) {
+    	if ((arg >= ABSMAXCOLS-1) || !growtbl(GROWCOL, 0, arg+1)) {
+	    error("You can't hide the last col");
+	    return;
 	}
     }
     FullUpdate++;
@@ -2023,6 +2155,7 @@ clearent(struct ent *v)
     v->format = (char *)0;
     v->flags |= (is_changed);
     v->flags &= ~(is_valid);
+    v->flags |= (is_cleared);
     changed++;
     modflg++;
 }
@@ -2037,9 +2170,10 @@ constant(register struct enode *e)
 	 e == (struct enode *)0
 	 || e->op == O_CONST
 	 || e->op == O_SCONST
+	 || e->op == 'm' && constant(e->e.o.right)
 	 || (
 	     e->op != O_VAR
-	     && (e->op & REDUCE) != REDUCE
+	     && !(e->op & REDUCE)
 	     && constant(e->e.o.left)
 	     && constant(e->e.o.right)
 	     && e->op != EXT	 /* functions look like constants but aren't */
@@ -2059,7 +2193,7 @@ efree(struct enode *e)
 {
     if (e) {
 	if (e->op != O_VAR && e->op !=O_CONST && e->op != O_SCONST
-		&& (e->op & REDUCE) != REDUCE) {
+		&& !(e->op & REDUCE)) {
 	    efree(e->e.o.left);
 	    efree(e->e.o.right);
 	}
@@ -2102,7 +2236,8 @@ decodev(struct ent_ptr v)
 {
 	struct range *r;
 
-	if (!v.vp) (void) sprintf(line+linelim,"VAR?");
+	if (!v.vp || v.vp->flags & is_deleted)
+	    (void) sprintf(line + linelim, "@ERR");
 	else if (!find_range((char *)0, 0, v.vp, v.vp, &r) && !r->r_is_range)
 	    (void) sprintf(line+linelim, "%s", r->r_name);
 	else {
@@ -2127,7 +2262,7 @@ coltoa(int col)
     }
     *p++ = col+'A';
     *p = '\0';
-    return(rname);
+    return (rname);
 }
 
 /*
@@ -2152,10 +2287,11 @@ decompile(register struct enode *e, int	priority)
 	int mypriority;
 	switch (e->op) {
 	default: mypriority = 99; break;
-	case '?': mypriority = 1; break;
-	case ':': mypriority = 2; break;
-	case '|': mypriority = 3; break;
-	case '&': mypriority = 4; break;
+	case ';': mypriority = 1; break;
+	case '?': mypriority = 2; break;
+	case ':': mypriority = 3; break;
+	case '|': mypriority = 4; break;
+	case '&': mypriority = 5; break;
 	case '<': case '=': case '>': mypriority = 6; break;
 	case '+': case '-': case '#': mypriority = 8; break;
 	case '*': case '/': case '%': mypriority = 10; break;
@@ -2187,10 +2323,12 @@ decompile(register struct enode *e, int	priority)
 	case REDUCE | 'a': range_arg( "@avg(", e); break;
 	case REDUCE | 'c': range_arg( "@count(", e); break;
 	case REDUCE | 's': range_arg( "@stddev(", e); break;
+	case REDUCE | 'R': range_arg( "@rows(", e); break;
+	case REDUCE | 'C': range_arg( "@cols(", e); break;
 	case REDUCE | MAX: range_arg( "@max(", e); break;
 	case REDUCE | MIN: range_arg( "@min(", e); break;
 
-	case ABS:		one_arg( "@abs(", e); break;
+	case ABS:	one_arg( "@abs(", e); break;
 	case ACOS:	one_arg( "@acos(", e); break;
 	case ASIN:	one_arg( "@asin(", e); break;
 	case ATAN:	one_arg( "@atan(", e); break;
@@ -2247,11 +2385,41 @@ decompile(register struct enode *e, int	priority)
 	case MYROW:	for (s = "@myrow"; (line[linelim++] = *s++); );
 			linelim--;
 			break;
-	case MYCOL:	for ( s = "@mycol"; (line[linelim++] = *s++); );
+	case MYCOL:	for (s = "@mycol"; (line[linelim++] = *s++); );
 			linelim--;
 			break;
 	case COLTOA:	one_arg( "@coltoa(", e); break;
-	case NUMITER:	for ( s = "@numiter"; (line[linelim++] = *s++); );
+	case NUMITER:	for (s = "@numiter"; (line[linelim++] = *s++); );
+                        linelim--;
+                        break;
+	case ERR_:	for (s = "@err"; (line[linelim++] = *s++); );
+                        linelim--;
+                        break;
+	case PI_:	for (s = "@pi"; (line[linelim++] = *s++); );
+                        linelim--;
+                        break;
+	case BLACK:	for (s = "@black"; (line[linelim++] = *s++); );
+                        linelim--;
+                        break;
+	case RED:	for (s = "@red"; (line[linelim++] = *s++); );
+                        linelim--;
+                        break;
+	case GREEN:	for (s = "@green"; (line[linelim++] = *s++); );
+                        linelim--;
+                        break;
+	case YELLOW:	for (s = "@yellow"; (line[linelim++] = *s++); );
+                        linelim--;
+                        break;
+	case BLUE:	for (s = "@blue"; (line[linelim++] = *s++); );
+                        linelim--;
+                        break;
+	case MAGENTA:	for (s = "@magenta"; (line[linelim++] = *s++); );
+                        linelim--;
+                        break;
+	case CYAN:	for (s = "@cyan"; (line[linelim++] = *s++); );
+                        linelim--;
+                        break;
+	case WHITE:	for (s = "@white"; (line[linelim++] = *s++); );
                         linelim--;
                         break;
 	default:	decompile(e->e.o.left, mypriority);
@@ -2377,10 +2545,10 @@ editv(int row, int col)
     p = lookat(row, col);
     (void) sprintf(line, "let %s = ", v_name(row, col));
     linelim = strlen(line);
-    if (p->flags & is_strexpr || p->expr == 0) {
+    if ((p->flags & is_strexpr || p->expr == NULL) && p->flags & is_valid) {
 	(void) sprintf(line+linelim, "%.15g", p->v);
 	linelim = strlen(line);
-    } else {
+    } else if (p->expr) {
         editexp(row, col);
     }
 }
