@@ -7,7 +7,7 @@
  *
  *              More mods Robert Bond, 12/86
  *		More mods by Alan Silverstein, 3/88, see list of changes.
- *		$Revision: 7.7 $
+ *		$Revision: 7.13 $
  *
  */
 
@@ -105,22 +105,27 @@ struct key statres[] = {
 int
 yylex()
 {
-    char *p = line+linelim;
+    char *p = line + linelim;
     int ret;
-    static bool isfunc = 0;
+    static int isfunc = 0;
+    static bool isgoto = 0;
     static bool colstate = 0;
     static int dateflag;
+    static char *tokenst = NULL;
+    static int tokenl;
 
     while (isspace(*p)) p++;
     if (*p == '\0') {
-	isfunc = 0;
+	isfunc = isgoto = 0;
 	ret = -1;
     } else if (isalpha(*p) || (*p == '_')) {
-	char *tokenst = p;
 	register char *la;	/* lookahead pointer */
-	int tokenl;
 	register struct key *tblp;
-	tokenl = 0;
+
+	if (!tokenst) {
+	    tokenst = p;
+	    tokenl = 0;
+	}
 	/*
 	 *  This picks up either 1 or 2 alpha characters (a column) or
 	 *  tokens made up of alphanumeric chars and '_' (a function or
@@ -137,7 +142,7 @@ yylex()
 	 * A COL is 1 or 2 char alpha with nothing but digits following
 	 * (no alpha or '_')
 	 */
-	if (tokenl && tokenl <= 2 && isfunc < 1 && (colstate ||
+	if (!isdigit(*tokenst) && tokenl && tokenl <= 2 && (colstate ||
 		(isdigit(*(la-1)) && !(isalpha(*la) || (*la == '_'))))) {
 	    ret = COL;
 	    yylval.ival = atocol(tokenst, tokenl);
@@ -158,6 +163,11 @@ yylex()
 			if (i >= tokenl) {
 			    ret = tblp->val;
 			    colstate = (ret <= S_FORMAT);
+			    if (isgoto) {
+				isfunc = isgoto = 0;
+				if (ret != K_ERROR && ret != K_INVALID)
+				    ret = WORD;
+			    }
 			    break;
 			}
 		    }
@@ -165,7 +175,8 @@ yylex()
 	    if (ret == WORD) {
 		struct range *r;
 		if (!find_range(tokenst, tokenl,
-			(struct ent *)0, (struct ent *)0, &r)) {
+			(struct ent *)0, (struct ent *)0, &r) &&
+			strlen(r->r_name) == tokenl) {
 		    yylval.rval.left = r->r_left;
 		    yylval.rval.right = r->r_right;
 		    if (r->r_is_range)
@@ -193,7 +204,8 @@ yylex()
 	    (void) signal(SIGFPE, sig_save);
 	    yylval.fval = v;
 	    error("Floating point exception\n");
-	    isfunc = 0;
+	    isfunc = isgoto = 0;
+	    tokenst = NULL;
 	    return FNUMBER;
 	}
 
@@ -202,8 +214,12 @@ yylex()
 	    dateflag--;
 	} else {
 	    if (*p != '.') {
-		do v = v*10.0 + (double) ((unsigned) *p - '0');
-		while (isdigit(*++p));
+		tokenst = p;
+		tokenl = 0;
+		do {
+		    v = v*10.0 + (double) ((unsigned) *p - '0');
+		    tokenl++;
+		} while (isdigit(*++p));
 		if (dateflag) {
 		    ret = NUMBER;
 		    yylval.ival = (int)v;
@@ -218,6 +234,9 @@ yylex()
 		    ret = NUMBER;
 		    yylval.ival = (int)v;
 		    dateflag = 2;
+		} else if (isalpha(*p) || *p == '_') {
+		    linelim = p - line;
+		    return (yylex());
 		}
 	    }
 	    if (!dateflag && (*p=='.' || *p == 'e' || *p == 'E')) {
@@ -250,7 +269,8 @@ yylex()
         ptr = scxmalloc((unsigned)(ptr-p));
 	yylval.sval = ptr;
 	p++;
-	while (*p && ((*p != '"') || (*(p-1) == '\\')))
+	while (*p && ((*p != '"') ||
+		(*(p-1) == '\\' && *(p+1) != '\0' && *(p+1) != '\n')))
 	    *ptr++ = *p++;
 	*ptr = '\0';
 	if (*p)
@@ -262,10 +282,13 @@ yylex()
 	if (*p)
 	    p++;
 	linelim = p-line;
+	tokenst = NULL;
 	return yylex();
     } else ret = *p++;
     linelim = p-line;
-    if (!isfunc) isfunc = ((ret == '@') - (ret == S_SET || ret == S_GOTO));
+    if (!isfunc) isfunc = ((ret == '@') + (ret == S_GOTO) - (ret == S_SET));
+    if (ret == S_GOTO) isgoto = TRUE;
+    tokenst = NULL;
     return ret;
 }
 
@@ -284,10 +307,10 @@ atocol(char *string, int len)
 {
     register int col;
 
-    col = (string[0] & 0137) - 'A';
+    col = (toupper(string[0])) - 'A';
 
     if (len == 2)		/* has second char */
-	col = ((col + 1) * 26) + ((string[1] & 0137) - 'A');
+	col = ((col + 1) * 26) + ((toupper(string[1])) - 'A');
 
     return (col);
 }
@@ -598,10 +621,10 @@ nmgetch()
 {
     register int c;
 
-    c = wgetch(stdscr);
+    c = getch();
     switch (c) {
 #ifdef KEY_SELECT
-    case KEY_SELECT:c = 'm'; break;
+	case KEY_SELECT:	c = 'm';	break;
 #endif
 #ifdef KEY_C1
 /* This stuff works for a wyse wy75 in ANSI mode under 5.3.  Good luck. */
