@@ -7,7 +7,7 @@
  *
  *              More mods Robert Bond, 12/86
  *		More mods by Alan Silverstein, 3-4/88, see list of changes.
- *		$Revision: 7.7 $
+ *		$Revision: 7.13 $
  *
  */
 
@@ -33,18 +33,34 @@ extern int VMS_read_raw;   /*sigh*/
 #define nonl()	 (_tty.sg_flags &= ~CRMOD, _pfast = TRUE, stty(_tty_ch, &_tty))
 #endif
 
-void	repaint(int x, int y, int len);
+#define SHOWROWS 2
+#define SHOWCOLS 3
+
+void	repaint(int x, int y, int len, int attron, int attroff);
 
 char	under_cursor = ' '; /* Data under the < cursor */
 char	mode_ind = 'i';
 extern	char    revmsg[];
 
+int	lines, cols;
 int	rows, lcols;
 int	lastmx, lastmy;	/* Screen address of the cursor */
-int	lastcol;	/* Spreadsheet Column the cursor was in last */
-int	rescol = 4;	/* columns reserved for row numbers */
+int	lastrow = -1;	/* Spreadsheet Row the cursor was in last */
+int	lastcol = -1;	/* Spreadsheet Column the cursor was in last */
+int	lastendrow = -1;	/* Last bottom row of screen */
+int	lastftoprows = 0;	/* Rows in top of frame cursor was in last */
+int	lastfbottomrows = 0;	/* Rows in bottom of frame cursor was in last */
+int	lastfleftcols = 0;	/* Columns in left side of frame cursor was
+				    in last */
+int	lastfrightcols = 0;
+struct	frange *lastfr = 0;	/* Last framed range we were in */
+bool	frTooLarge = 0;	/* If set, either too many rows or too many columns
+			    exist in frame to allow room for the scrolling
+			    portion of the framed range */
+int	framerows;	/* Rows in current frame */
+int	framecols;	/* Columns in current frame */
+int	rescol = 4;	/* Columns reserved for row numbers */
 extern	int *fwidth;
-extern	int showrange;	/* Causes ranges to be highlighted	*/
 extern	int showneed;	/* Causes cells needing values to be highlighted */
 extern	int showexpr;	/* Causes cell exprs to be displayed, highlighted */
 extern	int shownote;	/* Causes cells with attached notes to be highlighted */
@@ -68,57 +84,54 @@ update(int anychanged)		/* did any cell really change in value? */
     struct ent			**pp;
     int				mxrow, mxcol;
     int				minsr = 0, minsc = 0, maxsr = 0, maxsc = 0;
-    int				r;
-    int				i;
-    static int			lastcurcol = -1, lastcurrow = -1;
+    int				r, i;
     struct frange		*fr;
-    static struct frange	*lastfr = 0;
+    struct crange		*cr;
     int				ftoprows, fbottomrows, fleftcols, frightcols;
     int				ftrows, fbrows, flcols, frcols;
-    static int			lastftoprows = 0;
-    static int			lastfleftcols = 0;
-    static bool			frTooLarge = 0;
 
 #ifndef MSDOS
     /*
      * If receiving input from a pipeline, don't display spreadsheet data
      * on screen.
      */
-    if (!isatty(STDIN_FILENO)) return;
+    if (!usecurses) return;
 #endif
 
+    getmaxyx(stdscr, lines, cols);
     fr = lastfr;
     if (!(fr && fr->or_left->row <= currow &&	/* If we've left the	    */
 	    fr->or_left->col <= curcol &&	/* previous framed range... */
 	    fr->or_right->row >= currow &&
 	    fr->or_right->col >= curcol)) {
 	fr = find_frange(currow, curcol);
-	FullUpdate++;
+	if (fr != lastfr)
+	    FullUpdate++;
     }
     if (fr) {
 	ftoprows = fr->ir_left->row - fr->or_left->row;
 	fbottomrows = fr->or_right->row - fr->ir_right->row;
 	fleftcols = fr->ir_left->col - fr->or_left->col;
 	frightcols = fr->or_right->col - fr->ir_right->col;
-	ftrows = RESROW + ftoprows + fbottomrows;
-	flcols = rescol;
+	framerows = RESROW + ftoprows + fbottomrows;
+	framecols = rescol;
 	for (r = fr->or_left->row - 1, i = ftoprows; i; i--)
 	    if (row_hidden[r+i])
-		ftrows--;
+		framerows--;
 	for (r = fr->or_right->row - 1, i = fbottomrows; i; i--)
 	    if (row_hidden[r-i])
-		ftrows--;
+		framerows--;
 	for (r = fr->or_left->col - 1, i = fleftcols; i; i--) {
 	    if (col_hidden[r+i])
 		continue;
-	    flcols += fwidth[r+i];
+	    framecols += fwidth[r+i];
 	}
 	for (r = fr->or_right->col - 1, i = frightcols; i; i--) {
 	    if (col_hidden[r-i])
 		continue;
-	    flcols += fwidth[r-i];
+	    framecols += fwidth[r-i];
 	}
-	if (ftrows >= LINES || flcols >= COLS) {
+	if (framerows >= lines || framecols >= cols) {
 	    frTooLarge = TRUE;
 	    if (FullUpdate) {
 		error("Frame too large for screen size - ignoring");
@@ -129,47 +142,43 @@ update(int anychanged)		/* did any cell really change in value? */
 	} else {
 	    frTooLarge = FALSE;
 	    if (strow >= fr->or_left->row) {
-		if (currow > fr->ir_left->row)
-		    ftoprows = fr->ir_left->row - fr->or_left->row;
-		else
-		    ftoprows = currow - fr->or_left->row;
-		if (strow < fr->or_left->row + ftoprows)
-		    strow = fr->or_left->row + ftoprows;
+		if (fr == lastfr && strow < fr->or_left->row + ftoprows)
+		    strow = fr->or_left->row;
 		else if (strow > fr->ir_right->row) {
-		    strow = fr->ir_right->row + 1;
+		    strow = fr->ir_right->row;
 		    FullUpdate++;
 		}
-	    } else
-		ftoprows = 0;
-	    if (stcol >= fr->or_left->col) {
-		if (curcol > fr->ir_left->col)
-		    fleftcols = fr->ir_left->col - fr->or_left->col;
-		else
-		    fleftcols = curcol - fr->or_left->col;
-		if (stcol < fr->or_left->col + fleftcols)
-		    stcol = fr->or_left->col + fleftcols;
-		else if (stcol > fr->ir_right->col) {
-		    stcol = fr->ir_right->col + 1;
-		    FullUpdate++;
-		}
-	    } else
-		fleftcols = 0;
-	    if (fr) {
-		if (currow < fr->ir_right->row)
-		    fbottomrows = fr->or_right->row - fr->ir_right->row;
-		else
-		    fbottomrows = fr->or_right->row - currow;
-		if (curcol < fr->ir_right->col)
-		    frightcols = fr->or_right->col - fr->ir_right->col;
-		else
-		    frightcols = fr->or_right->col - curcol;
 	    }
+	    if (stcol >= fr->or_left->col) {
+		if (stcol < fr->or_left->col + fleftcols)
+		    stcol = fr->or_left->col;
+		else if (stcol > fr->ir_right->col) {
+		    stcol = fr->ir_right->col;
+		    FullUpdate++;
+		}
+	    }
+	    if (fr == lastfr && currow == lastrow)
+		fbottomrows = lastfbottomrows;
+	    else if (currow < fr->ir_right->row)
+		fbottomrows = fr->or_right->row - fr->ir_right->row;
+	    else
+		fbottomrows = fr->or_right->row - currow;
+	    if (fr == lastfr && curcol == lastcol)
+		frightcols = lastfrightcols;
+	    else if (curcol < fr->ir_right->col)
+		frightcols = fr->or_right->col - fr->ir_right->col;
+	    else
+		frightcols = fr->or_right->col - curcol;
 	}
-    } else
+    } else {
 	ftoprows = fbottomrows = fleftcols = frightcols = 0;
-    if (fr != lastfr && !gs.stflag) {
-	strow -= lastftoprows;
-	stcol -= lastfleftcols;
+	framerows = framecols = 0;
+    }
+    if (fr != lastfr && !gs.stflag && lastfr) {
+	if (strow >= lastfr->ir_left->row)
+	    strow -= lastftoprows;
+	if (stcol >= lastfr->ir_left->col)
+	    stcol -= lastfleftcols;
     }
 
     ftrows = ftoprows;
@@ -198,32 +207,41 @@ update(int anychanged)		/* did any cell really change in value? */
      * Place the cursor on the screen.  Set col, curcol, stcol, lastcol as
      * needed.  If strow and stcol are negative, centering is forced.
      */
-    if ((curcol != lastcurcol) || FullUpdate) {
+    if ((curcol != lastcol) || FullUpdate) {
 	while (col_hidden[curcol])   /* You can't hide the last row or col */
 	    curcol++;
-	if (fwidth[curcol] > COLS - rescol - 2) {
+	if (fwidth[curcol] > cols - rescol - 2) {
 	    error("column %s too wide - resizing", coltoa(curcol));
-	    doformat(curcol, curcol, COLS - rescol - 2,
+	    doformat(curcol, curcol, cols - rescol - 2,
 		    precision[curcol], realfmt[curcol]);
 	}
 
 	/* First see if the last display still covers curcol */
-	if (stcol >= 0 && stcol <= curcol) { 
-	    if (fr && fr != lastfr) {
-		if (stcol >= fr->or_left->col &&
-			stcol < fr->or_left->col + fleftcols)
-		    stcol = fr->or_left->col + fleftcols;
-		else if (stcol < fr->or_left->col) {
-		    fleftcols = 0;
-		    flcols = 0;
-		} else if (!gs.stflag) {
+	if (stcol >= 0 && stcol <= curcol) {
+	    int c = 0;
+
+	    if (fr) {
+		if (fr != lastfr) {
+		    if (stcol == fr->or_left->col)
+			stcol += fleftcols;
+		    else if (stcol >= fr->or_left->col && !gs.stflag) {
+			stcol += fleftcols;
+			if (stcol > fr->ir_right->col)
+			    stcol = fr->ir_right->col + 1;
+		    }
+		} else if (stcol == fr->or_left->col)
 		    stcol += fleftcols;
-		    if (stcol > fr->ir_right->col)
-			stcol = fr->ir_right->col + 1;
-		}
 	    }
-	    for (i = stcol, lcols = 0, col = rescol + flcols + frcols;
-		    (col + fwidth[i]) < COLS-1 && i < maxcols; i++) {
+	    i = stcol;
+	    lcols = 0;
+	    col = rescol + frcols;
+	    if (fr && stcol >= fr->or_left->col)
+		if (stcol < fr->ir_left->col)
+		    i = fr->or_left->col;
+		else
+		    col += flcols;
+	    for (; (col + fwidth[i] < cols-1 || col_hidden[i] || i < curcol) &&
+		    i < maxcols; i++) {
 		lcols++;
 		if (fr && i == fr->ir_right->col + 1) {
 		    col -= frcols;
@@ -231,8 +249,34 @@ update(int anychanged)		/* did any cell really change in value? */
 		}
 		if (col_hidden[i])
 		    continue;
+
+		/* If there isn't room for more columns, and we haven't yet
+		 * reached the current column, start removing columns from
+		 * the left.
+		 */
+		while (col + fwidth[i] > cols - 2) {
+		    lcols--;
+		    col -= fwidth[stcol];
+		    while (col_hidden[++stcol]) /**/ ;
+		    FullUpdate++;
+		    c++;
+		}
 		col += fwidth[i];
 	    }
+	    if (!frTooLarge && fr && curcol <= stcol + lcols &&
+		    fr->ir_left->col >= stcol + lcols) {
+		while (stcol + lcols < fr->ir_left->col) {
+		    col -= fwidth[stcol];
+		    lcols--;
+		    while (col_hidden[++stcol])
+			lcols--;
+		    while (col + fwidth[stcol + lcols] < cols - 1) {
+			col += fwidth[stcol + lcols];
+			lcols++;
+		    }
+		}
+	    } else if (c)
+		stcol = -1;
 	}
 
 	while (stcol < 0 || curcol < stcol || stcol + lcols - 1 < curcol ||
@@ -251,17 +295,20 @@ update(int anychanged)		/* did any cell really change in value? */
 		    curcol <= fr->ir_left->col && stcol < curcol &&
 		    curcol <= stcol + lcols + fr->ir_left->col -
 			    fr->or_left->col) {
-		while (stcol + lcols < fr->ir_left->col && !frTooLarge)
+		while ((stcol + lcols < fr->ir_left->col && !frTooLarge) ||
+			(colsinrange != fwidth[curcol] && stcol != curcol &&
+			stcol + lcols - 1 < gs.g_lastcol)) {
 		    if (col_hidden[++stcol]) lcols--;
+		}
 	    } else {
 		/* Try to put the cursor in the center of the screen.
 		 * If we've just jumped to a range using the goto command,
 		 * center the range instead.
 		 */
-		colsinrange = (colsinrange > COLS - rescol -
+		colsinrange = (colsinrange > cols - rescol -
 			flcols - frcols - 2 ?
-			COLS - rescol - flcols - frcols - 2 : colsinrange);
-		col = (COLS - rescol - flcols - frcols - colsinrange)/2; 
+			cols - rescol - flcols - frcols - 2 : colsinrange);
+		col = (cols - rescol - flcols - frcols - colsinrange)/2; 
 		stcol = curcol;
 		for (i = curcol - 1;
 			i >= (fr ? fr->or_left->col + fleftcols : 0) &&
@@ -272,12 +319,23 @@ update(int anychanged)		/* did any cell really change in value? */
 			continue;
 		    col -= fwidth[i];
 		}
-		if (fr && stcol < fr->or_left->col + fleftcols)
+		if (fr && stcol < fr->or_left->col + fleftcols) {
 		    stcol = fr->or_left->col + fleftcols;
+		    if (curcol < stcol)
+			stcol = curcol;
+		}
 	    }
 	    /* Now pick up the counts again */
-	    for (i = stcol, lcols = 0, col = rescol + flcols + frcols;
-		    (col + fwidth[i]) < COLS-1 && i < maxcols; i++) {
+	    i = stcol;
+	    lcols = 0;
+	    col = rescol + frcols;
+	    if (fr && stcol >= fr->or_left->col)
+		if (stcol < fr->ir_left->col)
+		    i = fr->or_left->col;
+		else
+		    col += flcols;
+	    for (; (col + fwidth[i] < cols-1 || col_hidden[i] || i < curcol) &&
+		    i < maxcols; i++) {
 		lcols++;
 		if (fr && i == fr->ir_right->col + 1) {
 		    col -= frcols;
@@ -285,32 +343,48 @@ update(int anychanged)		/* did any cell really change in value? */
 		}
 		if (col_hidden[i])
 		    continue;
+
 		col += fwidth[i];
 	    }
 	}
-	lastcurcol = curcol;
+    }
+    if (fleftcols && stcol >= fr->or_left->col &&
+	    stcol < fr->or_left->col + fleftcols) {
+	lcols += (fr->or_left->col - stcol);
+	stcol = fr->or_left->col + fleftcols;
+	if (curcol < stcol)
+	    stcol = curcol;
     }
 
     /* Now - same process on the rows as the columns */
-    if ((currow != lastcurrow) || FullUpdate) {
+    if ((currow != lastrow) || FullUpdate) {
 	while (row_hidden[currow])   /* You can't hide the last row or col */
 	    currow++;
-	if (strow >= 0 && strow <= currow) { 
-	    if (fr && fr != lastfr) {
-		if (strow >= fr->or_left->row &&
-			strow < fr->or_left->row + ftoprows)
-		    strow = fr->or_left->row + ftoprows;
-		else if (strow < fr->or_left->row) {
-		    ftoprows = 0;
-		    ftrows = 0;
-		} else if (!gs.stflag) {
+	if (strow >= 0 && strow <= currow) {
+	    int c = 0;
+
+	    if (fr) {
+		if (fr != lastfr) {
+		    if (strow == fr->or_left->row)
+			strow += ftoprows;
+		    else if (strow >= fr->or_left->row && !gs.stflag) {
+			strow += ftoprows;
+			if (strow > fr->ir_right->row)
+			    strow = fr->ir_right->row + 1;
+		    }
+		} else if (strow == fr->or_left->row)
 		    strow += ftoprows;
-		    if (strow > fr->ir_right->row)
-			strow = fr->ir_right->row + 1;
-		}
 	    }
-	    for (i = strow, rows = 0, row = RESROW + ftrows + fbrows;
-		    row < LINES && i < maxrows; i++) {
+	    i = strow;
+	    rows = 0;
+	    row = RESROW + fbrows;
+	    if (fr && strow >= fr->or_left->row)
+		if (strow < fr->ir_left->row)
+		    i = fr->or_left->row;
+		else
+		    row += ftrows;
+	    for (; (row < lines || row_hidden[i] || i < currow) && i < maxrows;
+		    i++) {
 		rows++;
 		if (fr && i == fr->ir_right->row + 1) {
 		    row -= fbrows;
@@ -318,13 +392,27 @@ update(int anychanged)		/* did any cell really change in value? */
 		}
 		if (row_hidden[i])
 		    continue;
+
+		/* If there isn't room for more rows, and we haven't yet
+		 * reached the current row, start removing rows from the
+		 * top.
+		 */
+		if (row >= lines) {
+		    rows--;
+		    row--;
+		    while (row_hidden[++strow]) /**/;
+		    FullUpdate++;
+		    c++;
+		}
 		row++;
 	    }
-	    if (fr && currow <= strow + rows &&
+	    if (!frTooLarge && fr && currow <= strow + rows &&
 		    fr->ir_left->row >= strow + rows) {
-		while (strow + rows < fr->ir_left->row && !frTooLarge)
-		    if (row_hidden[++strow]) rows--;
-	    }
+		while (strow + rows < fr->ir_left->row) {
+		    while (row_hidden[++strow]) /**/;
+		}
+	    } else if (c && currow > lastendrow)
+		strow = -1;
 	}
 
 	while (strow < 0 || currow < strow || strow + rows - 1 < currow ||
@@ -342,20 +430,23 @@ update(int anychanged)		/* did any cell really change in value? */
 		    currow <= fr->ir_left->row && strow < currow &&
 		    currow <= strow + rows + fr->ir_left->row -
 			    fr->or_left->row) {
-		while (strow + rows < fr->ir_left->row && !frTooLarge)
+		while ((strow + rows < fr->ir_left->row && !frTooLarge) ||
+			(rowsinrange > 1 && strow != currow &&
+			strow + rows - 1 < gs.g_lastrow)) {
 		    if (row_hidden[++strow]) rows--;
+		}
 	    } else {
 		/* Try to put the cursor in the center of the screen.
 		 * If we've just jumped to a range using the goto command,
 		 * center the range instead.
 		 */
-		rowsinrange = (rowsinrange > LINES - RESROW - ftrows - fbrows ?
-			LINES - RESROW : rowsinrange);
-		row = (LINES - RESROW - ftrows - fbrows - rowsinrange)/2; 
+		rowsinrange = (rowsinrange > lines - RESROW - ftrows - fbrows ?
+			lines - RESROW - ftrows - fbrows : rowsinrange);
+		row = (lines - RESROW - ftrows - fbrows - rowsinrange)/2; 
 		strow = currow;
 		for (i = currow - 1;
 			i >= (fr ? fr->or_left->row + ftoprows : 0) &&
-			row > 0; i--) {
+			(row > 0 || row_hidden[i]); i--) {
 		    strow--;
 		    if (row_hidden[i])
 			continue;
@@ -363,10 +454,20 @@ update(int anychanged)		/* did any cell really change in value? */
 		}
 		if (fr && strow < fr->or_left->row + ftoprows)
 		    strow = fr->or_left->row + ftoprows;
+		    if (currow < strow)
+			strow = currow;
 	    }
 	    /* Now pick up the counts again */
-	    for (i = strow, rows = 0, row = RESROW + ftrows + fbrows;
-		    row < LINES && i < maxrows; i++) {
+	    i = strow;
+	    rows = 0;
+	    row = RESROW + fbrows;
+	    if (fr && strow >= fr->or_left->row)
+		if (strow < fr->ir_left->row)
+		    i = fr->or_left->row;
+		else
+		    row += ftrows;
+	    for (; (row < lines || row_hidden[i] || i < currow) && i < maxrows;
+		    i++) {
 		rows++;
 		if (fr && i == fr->ir_right->row + 1) {
 		    row -= fbrows;
@@ -374,56 +475,109 @@ update(int anychanged)		/* did any cell really change in value? */
 		}
 		if (row_hidden[i])
 		    continue;
+
 		row++;
 	    }
 	}
-	lastcurrow = currow;
     }
+    if (ftoprows && strow >= fr->or_left->row &&
+	    strow < fr->or_left->row + ftoprows) {
+	rows += (fr->or_left->row - strow);
+	strow = fr->or_left->row + ftoprows;
+	if (currow < strow)
+	    strow = currow;
+    }
+
     mxcol = frightcols ? fr->or_right->col : stcol + lcols - 1;
     mxrow = fbottomrows ? fr->or_right->row : strow + rows - 1;
     gs.stflag = 0;
     lastfr = fr;
     lastftoprows = ftoprows;
+    lastfbottomrows = fbottomrows;
     lastfleftcols = fleftcols;
+    lastfrightcols = frightcols;
 
     /* Get rid of cursor standout on the cell at previous cursor position */
     if (!FullUpdate) {
-	if (showcell)
-	    repaint(lastmx, lastmy, fwidth[lastcol]);
+	if (showcell) {
+	    pp = ATBL(tbl, lastrow, lastcol);
+	    if (color && has_colors()) {
+		if ((cr = find_crange(lastrow, lastcol)))
+		    color_set(cr->r_color, NULL);
+		else
+		    color_set(1, NULL);
+		if (*pp) {
+		    if (colorneg && (*pp)->flags & is_valid && (*pp)->v < 0) {
+			if (cr)
+			    color_set(((cr->r_color) % CPAIRS) + 1, NULL);
+			else
+			    color_set(2, NULL);
+		    }
+		    else if (colorerr && (*pp)->cellerror)
+			color_set(3, NULL);
+		}
+	    }
+	    repaint(lastmx, lastmy, fwidth[lastcol], 0, A_STANDOUT);
+	}
 
 	(void) move(lastmy, lastmx+fwidth[lastcol]);
 
-	if ((inch() & A_CHARTEXT ) == '<')
-	    (void) addch(under_cursor);
+	if ((inch() & A_CHARTEXT) == '<')
+	    (void) addch(under_cursor | (inch() & A_ATTRIBUTES));
 
-	standout();
-	repaint(lastmx, RESROW - 1, fwidth[lastcol]);
-	repaint(0, lastmy, rescol - 1);
-	standend();
+	repaint(lastmx, RESROW - 1, fwidth[lastcol], A_STANDOUT, 0);
+	repaint(0, lastmy, rescol - 1, A_STANDOUT, 0);
+	if (color && has_colors())
+	    color_set(1, NULL);
     }
+    lastrow = currow;
+    lastcol = curcol;
+    lastendrow = strow + rows;
 
     /* where is the the cursor now? */
-    lastmy =  RESROW + ftrows;
-    for (row = strow; row < currow; row++)
+    lastmy =  RESROW;
+    if (fr && strow >= fr->or_left->row)
+	if (strow < fr->ir_left->row)
+	    row = fr->or_left->row;
+	else {
+	    row = strow;
+	    lastmy += ftrows;
+	}
+    else
+	row = strow;
+    for (; row < currow; row++)
 	if (!row_hidden[row])
 	    lastmy++;
 
-    lastmx = rescol + flcols;
-    for (col = stcol; col < curcol; col++)
+    lastmx = rescol;
+    if (fr && stcol >= fr->or_left->col)
+	if (stcol < fr->ir_left->col)
+	    col = fr->or_left->col;
+	else {
+	    col = stcol;
+	    lastmx += flcols;
+	}
+    else
+	col = stcol;
+    for (; col < curcol; col++)
 	if (!col_hidden[col])
 	    lastmx += fwidth[col];
-    lastcol = curcol;
+
+    if (color && has_colors())
+	color_set(1, NULL);
 
     if (FullUpdate || standlast) {
 	(void) move(2, 0);
 	(void) clrtobot();
 	(void) standout();
 
-	for (row = RESROW, i = (ftoprows ? fr->or_left->row : strow);
-		i <= (fbottomrows ? fr->or_right->row : mxrow); i++) {
-	    if (ftoprows && row == RESROW + ftrows)
-		i = strow;
-	    if (fbottomrows && row == LINES - fbrows)
+	for (row = RESROW, i = (ftoprows && strow >= fr->or_left->row ?
+		fr->or_left->row : strow);
+		i <= mxrow; i++) {
+	    if (ftoprows && strow >= fr->or_left->row &&
+		    row == RESROW + ftrows)
+		i = (strow < i ? i : strow);
+	    if (fbottomrows && row == lines - fbrows)
 		i = fr->or_right->row - fbottomrows + 1;
 	    if (row_hidden[i]) 
 		continue;
@@ -440,18 +594,20 @@ update(int anychanged)		/* did any cell really change in value? */
 	(void) move(2, 0);
 	(void) printw("%*s", rescol, " ");
 
-	for (col = rescol, i = (fleftcols ? fr->or_left->col : stcol);
-		i <= (frightcols ? fr->or_right->col : mxcol); i++) {
+	for (col = rescol, i = (fleftcols && stcol >= fr->or_left->col ?
+		fr->or_left->col : stcol);
+		i <= mxcol; i++) {
 	    register int k;
-	    if (fleftcols && col == rescol + flcols)
-		i = stcol;
-	    if (frightcols && col + fwidth[i] >= COLS - 1 - frcols &&
+	    if (fleftcols && stcol >= fr->or_left->col &&
+		    col == rescol + flcols)
+		i = (stcol < i ? i : stcol);
+	    if (frightcols && col + fwidth[i] >= cols - 1 - frcols &&
 		    i < fr->or_right->col - frightcols + 1)
 		i = fr->or_right->col - frightcols + 1;
 	    if (col_hidden[i])
 		continue;
 	    (void) move(2, col);
-	    k = fwidth[i]/2;
+	    k = (fwidth[i] - strlen(coltoa(i)))/2;
 	    if (k == 0)
 		(void) printw("%1s", coltoa(i));
 	    else
@@ -462,16 +618,45 @@ update(int anychanged)		/* did any cell really change in value? */
     }
 
     if (showrange) {
-	minsr = showsr < currow ? showsr : currow;
-	minsc = showsc < curcol ? showsc : curcol;
-	maxsr = showsr > currow ? showsr : currow;
-	maxsc = showsc > curcol ? showsc : curcol;
+	if (showrange == SHOWROWS) {
+	    minsr = showsr < currow ? showsr : currow;
+	    minsc = fr ? fr->or_left->col : 0;
+	    maxsr = showsr > currow ? showsr : currow;
+	    maxsc = fr ? fr->or_right->col : maxcols;
 
-	if (showtop) {
-	    (void) move(1,0);
-	    (void) clrtoeol();
-	    (void) printw("Default range:  %s",
+	    if (showtop) {
+		(void) move(1,0);
+		(void) clrtoeol();
+		(void) printw("Default range:  %d:%d", minsr, maxsr);
+	    }
+	} else if (showrange == SHOWCOLS) {
+	    minsr = 0;
+	    minsc = showsc < curcol ? showsc : curcol;
+	    maxsr = maxrows;
+	    maxsc = showsc > curcol ? showsc : curcol;
+
+	    if (showtop) {
+		char r[6];
+
+		strcpy(r, coltoa(minsc));
+		strcat(r, ":");
+		strcat(r, coltoa(maxsc));
+		(void) move(1,0);
+		(void) clrtoeol();
+		(void) printw("Default range:  %s", r);
+	    }
+	} else {
+	    minsr = showsr < currow ? showsr : currow;
+	    minsc = showsc < curcol ? showsc : curcol;
+	    maxsr = showsr > currow ? showsr : currow;
+	    maxsc = showsc > curcol ? showsc : curcol;
+
+	    if (showtop) {
+		(void) move(1,0);
+		(void) clrtoeol();
+		(void) printw("Default range:  %s",
 			    r_name(minsr, minsc, maxsr, maxsc));
+	    }
 	}
     }
 
@@ -484,8 +669,9 @@ update(int anychanged)		/* did any cell really change in value? */
 	standlast = FALSE;
       }
 
-      for (row = (ftoprows ? fr->or_left->row : strow), r = RESROW;
-	    row <= (fbottomrows ? fr->or_right->row : mxrow); row++) {
+      for (row = (ftoprows && strow >= fr->or_left->row ?
+		  fr->or_left->row : strow), r = RESROW;
+	    row <= mxrow; row++) {
 	int c = rescol;
 	int do_stand = 0;
 	int fieldlen;
@@ -493,28 +679,30 @@ update(int anychanged)		/* did any cell really change in value? */
 
 	if (row_hidden[row])
 	    continue;
-	if (ftoprows && r == RESROW + ftrows)
-	    row = strow;
-	if (fbottomrows && r == LINES - fbrows)
+	if (ftoprows && strow >= fr->or_left->row && r == RESROW + ftrows)
+	    row = (strow < row ? row : strow);
+	if (fbottomrows && r == lines - fbrows)
 	    row = fr->or_right->row - fbottomrows + 1;
-	for (pp = ATBL(tbl, row, col = (fleftcols ? fr->or_left->col : stcol));
+	for (pp = ATBL(tbl, row, col = (fleftcols && stcol >= fr->or_left->col ?
+		fr->or_left->col : stcol));
 		col <= mxcol;
 		pp += nextcol - col,  col = nextcol, c += fieldlen) {
 
-	    if (col_hidden[col]) {
-		fieldlen = 0;
-		continue;
-	    }
-	    if (fleftcols && c == rescol + flcols) {
-		col = stcol;
+	    if (fleftcols && stcol >= fr->or_left->col &&
+		    c == rescol + flcols) {
+		col = (stcol < col ? col : stcol);
 		pp = ATBL(tbl, row, col);
 	    }
-	    if (frightcols && c + fwidth[col] >= COLS - 1 - frcols &&
+	    if (frightcols && c + fwidth[col] >= cols - 1 - frcols &&
 		    col < fr->or_right->col - frightcols + 1) {
 		col = fr->or_right->col - frightcols + 1;
 		pp = ATBL(tbl, row, col);
 	    }
 	    nextcol = col + 1;
+	    if (col_hidden[col]) {
+		fieldlen = 0;
+		continue;
+	    }
 
 	    fieldlen = fwidth[col];
 
@@ -540,16 +728,23 @@ update(int anychanged)		/* did any cell really change in value? */
 
 		(void) move(r, c);
 		(void) standout();
+		if (color && has_colors() && (cr = find_crange(row, col)))
+		    color_set(cr->r_color, NULL);
 		standlast++;
 		if (!*pp) {	/* no cell, but standing out */
 		    (void) printw("%*s", fwidth[col], " ");
 		    (void) standend();
+		    if (color && has_colors())
+			color_set(1, NULL);
 		    continue;
 		} else
 		    do_stand = 1;
 	    } else
 		do_stand = 0;
 
+	    if ((cr = find_crange(row, col)) && color && has_colors())
+		color_set(cr->r_color, NULL);
+
 	    if ((*pp) && (((*pp)->flags & is_changed || FullUpdate) ||
 		    do_stand)) {
 		if (do_stand) {
@@ -563,10 +758,12 @@ update(int anychanged)		/* did any cell really change in value? */
 		 * Show expression; takes priority over other displays:
 		 */
 
-		if ((*pp)->cellerror)
+		if ((*pp)->cellerror) {
+		    if (color && colorerr && has_colors())
+			color_set(3, NULL);
 		    (void) printw("%*.*s", fwidth[col], fwidth[col],
 			(*pp)->cellerror == CELLERROR ? "ERROR" : "INVALID");
-		else
+		} else
 		if (showexpr && ((*pp)->expr)) {
 		    linelim = 0;
 		    editexp(row, col);		/* set line to expr */
@@ -581,28 +778,72 @@ update(int anychanged)		/* did any cell really change in value? */
 
 		    if ((*pp)->flags & is_valid) {
 			char field[FBUFLEN];
+			char *cfmt;
+			int note;
 
-			if ((*pp)->format) {
-			    if (*((*pp)->format) == ctl('d')) {
+			note = (*pp)->nrow >= 0 ? 1 : 0;
+			cfmt = (*pp)->format ? (*pp)->format :
+			    (realfmt[col] >= 0 && realfmt[col] < COLFORMATS &&
+			    colformat[realfmt[col]]) ?
+			    colformat[realfmt[col]] : NULL;
+			if (color && has_colors() && colorneg && (*pp)->v < 0) {
+			    if (cr)
+				color_set(((cr->r_color) % CPAIRS) + 1, NULL);
+			    else
+				color_set(2, NULL);
+			}
+			if (cfmt) {
+			    if (*cfmt == ctl('d')) {
 				time_t v = (time_t) ((*pp)->v);
 				strftime(field, sizeof(field),
-					((*pp)->format)+1, localtime(&v));
+					cfmt + 1, localtime(&v));
 			    } else
-			    (void) format((*pp)->format, (*pp)->v,
-				    field, sizeof(field));
+				(void) format(cfmt, precision[col], (*pp)->v,
+					field, sizeof(field));
 			} else {
-			    (void) engformat(realfmt[col], fwidth[col], 
+			    (void) engformat(realfmt[col], fwidth[col] - note,
 				precision[col], (*pp)->v, 
 				field, sizeof(field));
 			}
 			if (strlen(field) > fwidth[col]) {
-			    for(i = 0; i<fwidth[col]; i++)
+			    for (i = 0; i < fwidth[col]; i++) {
+				if (note) {
+				    attr_t attr;
+				    short curcolor;
+				    if (!i && color && has_colors()) {
+					attr_get(&attr, &curcolor, NULL);
+					color_set(4, NULL);
+				    }
+				    (void)addch('*');
+				    i++;
+				    if (!i && color && has_colors())
+					color_set(curcolor, NULL);
+				}
 				(void)addch('*');
+			    }
 			} else {
-			    if ((*pp)->format && *((*pp)->format) != ctl('d'))
-			    for(i = 0; i < fwidth[col] - strlen(field);i++)
-				(void)addch(' ');
+			    if (cfmt && *cfmt != ctl('d'))
+				for (i = 0;
+					i < fwidth[col] - strlen(field) - note;
+					i++)
+				    (void)addch(' ');
+			    if (note) {
+				attr_t attr;
+				short curcolor;
+				if (color && has_colors()) {
+				    attr_get(&attr, &curcolor, NULL);
+				    color_set(4, NULL);
+				}
+				(void)addch('*');
+				if (color && has_colors())
+				    color_set(curcolor, NULL);
+			    }
 			    (void)addstr(field);
+			    if (cfmt && *cfmt == ctl('d'))
+				for (i = 0;
+					i < fwidth[col] - strlen(field) - note;
+					i++)
+				    (void)addch(' ');
 			}
 		    }
 
@@ -616,19 +857,26 @@ update(int anychanged)		/* did any cell really change in value? */
 				    (*pp)->flags & is_valid,
 				    row, col, &nextcol, mxcol, &fieldlen,
 				    r, c, fr, frightcols, flcols, frcols);
-		    }
-		    else	/* repaint a blank cell: */
-		    if ((do_stand || !FullUpdate) &&
-				((*pp)->flags & is_changed) &&
-				!((*pp)->flags & is_valid) && !(*pp)->label) {
+		    } else	/* repaint a blank cell: */
+		    if ((((do_stand || !FullUpdate) &&
+			    ((*pp)->flags & is_changed)) ||
+			    (color && has_colors() &&
+			    cr && cr->r_color != 1)) &&
+			    !((*pp)->flags & is_valid) && !(*pp)->label) {
 			(void) printw("%*s", fwidth[col], " ");
 		    }
 		} /* else */
-
-		if (do_stand) {
-		    (void) standend();
-		    do_stand = 0;
-		}
+	    }
+	    else if (!*pp && color && has_colors && cr && cr->r_color != 1) {
+		move(r, c);
+		color_set(cr->r_color, NULL);
+		printw("%*s", fwidth[col], " ");
+	    }
+	    if (color && has_colors())
+		color_set(1, NULL);
+	    if (do_stand) {
+		(void) standend();
+		do_stand = 0;
 	    }
 	}
 	r++;
@@ -638,21 +886,38 @@ update(int anychanged)		/* did any cell really change in value? */
     /* place 'cursor marker' */
     if (showcell && (!showneed) && (!showexpr) && (!shownote)) {
 	(void) move(lastmy, lastmx);
-        (void) standout();
-        repaint(lastmx, lastmy, fwidth[lastcol]);
-        (void) standend();
+	pp = ATBL(tbl, currow, curcol);
+	if (color && has_colors()) {
+	    if ((cr = find_crange(currow, curcol)))
+		color_set(cr->r_color, NULL);
+	    else
+		color_set(1, NULL);
+	    if (*pp) {
+		if (colorneg && (*pp)->flags & is_valid && (*pp)->v < 0) {
+		    if (cr)
+			color_set(((cr->r_color) % CPAIRS) + 1, NULL);
+		    else
+			color_set(2, NULL);
+		} else if (colorerr && (*pp)->cellerror)
+		    color_set(3, NULL);
+	    }
+	}
+        repaint(lastmx, lastmy, fwidth[lastcol], A_STANDOUT, 0);
+	if (color && has_colors())
+	    color_set(1, NULL);
     }
 
-    repaint(lastmx, RESROW - 1, fwidth[lastcol]);
-    repaint(0, lastmy, rescol - 1);
+    repaint(lastmx, RESROW - 1, fwidth[lastcol], 0, A_STANDOUT);
+    repaint(0, lastmy, rescol - 1, 0, A_STANDOUT);
 
     (void) move(lastmy, lastmx+fwidth[lastcol]);
     under_cursor = (inch() & A_CHARTEXT);
-    if (! showcell)
-	(void) addch('<');
+    if (!showcell)
+	(void) addch('<' | (inch() & A_ATTRIBUTES));
 
     (void) move(0, 0);
     (void) clrtoeol();
+
     if (linelim >= 0) {
 	int ctlchars;
 
@@ -662,7 +927,7 @@ update(int anychanged)		/* did any cell really change in value? */
 	(void) addch(mode_ind);
 	(void) addstr("> ");
 	(void) addstr(line);
-	(void) move((linelim+3+ctlchars)/COLS, (linelim+3+ctlchars)%COLS);
+	(void) move((linelim+3+ctlchars)/cols, (linelim+3+ctlchars)%cols);
     } else {
 	if (showtop) {			/* show top line */
 	    register struct ent *p1;
@@ -736,18 +1001,21 @@ update(int anychanged)		/* did any cell really change in value? */
 		(void) addstr(" locked");
 	}
 	if (showcell)
-	    move(LINES - 1, COLS - 1);
+	    move(lines - 1, cols - 1);
 	else
 	    (void) move(lastmy, lastmx+fwidth[lastcol]);
     }
 
+    if (color && has_colors())
+	color_set(1, NULL);
+
     if (revmsg[0]) {
 	(void) move(0, 0);
 	(void) clrtoeol();	/* get rid of topline display */
 	(void) printw(revmsg);
 	*revmsg = '\0';		/* don't show it again */
 	if (showcell)
-	    move(LINES - 1, COLS - 1);
+	    move(lines - 1, cols - 1);
 	else
 	    (void) move(lastmy, lastmx + fwidth[lastcol]);
     }
@@ -757,14 +1025,11 @@ update(int anychanged)		/* did any cell really change in value? */
 
 /* redraw what is under the cursor from curses' idea of the screen */
 void
-repaint(int x, int y, int len)
+repaint(int x, int y, int len, int attron, int attroff)
 {
-    int c;
-
-    while(len-- > 0) {
+    while (len-- > 0) {
 	(void) move(y, x);
-	c = inch() & A_CHARTEXT;
-	(void) addch(c);
+	addch((inch() | attron) & ~attroff);
 	x++;
     }
 }
@@ -775,11 +1040,15 @@ int seenerr;
 void
 yyerror(char *err)
 {
-    if (seenerr) return;
-    seenerr++;
-    (void) move(1,0);
-    (void) clrtoeol();
-    (void) printw("%s: %.*s<=%s",err,linelim,line,line+linelim);
+    if (usecurses) {
+	if (seenerr) return;
+	seenerr++;
+	(void) move(1,0);
+	(void) clrtoeol();
+	(void) printw("%s: %.*s<=%s", err, linelim, line, line + linelim);
+    } else
+	(void) fprintf(stderr, "%s: %.*s<=%s\n", err, linelim, line,
+		       line + linelim);
 }
 
 #ifdef XENIX2_3
@@ -794,13 +1063,14 @@ startdisp()
     fd = dup(0);
 #endif
 #ifndef MSDOS
-    if (isatty(STDIN_FILENO)) {
+    if (usecurses) {
 #endif
+	int i;
 #ifdef TIOCGSIZE
 	{   struct ttysize size;
 	    if (ioctl(0, TIOCGSIZE, &size) == 0) { 
-		LINES = size.ts_lines;
-		COLS = size.ts_cols;
+		lines = size.ts_lines;
+		cols = size.ts_cols;
 	    }
 	}
 #endif
@@ -809,6 +1079,12 @@ startdisp()
 	(void) ioctl(fileno(stdin), TCGETA, & tmio);
 #endif
 	(void) initscr();
+	start_color();
+	for (i = 0; i < 8; i++)
+	    if (cpairs[i])
+		init_pair(i + 1, cpairs[i]->fg, cpairs[i]->bg);
+	if (color && has_colors())
+	    bkgdset(COLOR_PAIR(1) | ' ');
 #if sun
 	close(0);
 	dup(fd);
@@ -860,9 +1136,9 @@ void
 stopdisp()
 {
 #ifndef MSDOS
-    if (isatty(STDIN_FILENO)) {
+    if (usecurses) {
 #endif
-	deraw();
+	deraw(1);
 	resetkbd();
 	endwin();
 #ifdef XENIX2_3
@@ -876,51 +1152,73 @@ stopdisp()
 /* init curses */
 #ifdef VMS
 
+void
 goraw()
 {
-    VMS_read_raw = 1;
-    FullUpdate++;
+    if (usecurses) {
+	VMS_read_raw = 1;
+	if (color && has_colors())
+	    bkgdset(COLOR_PAIR(1) | ' ');
+	FullUpdate++;
+    }
 }
 
-deraw()
+void
+deraw(int ClearLastLine)
 {
-    (void) move (LINES - 1, 0);
-    (void) clrtoeol();
-    (void) refresh();
-    VMS_read_raw = 0;
+    if (usecurses) {
+	if (ClearLastLine) {
+	    if (color && has_colors())
+		bkgdset(COLOR_PAIR(0) | ' ');
+	    (void) move(lines - 1, 0);
+	    (void) clrtoeol();
+	    (void) refresh();
+	}
+	VMS_read_raw = 0;
+    }
 }
 
 #else /* VMS */
 void
 goraw()
 {
+    if (usecurses) {
 #if SYSV2 || SYSV3
-    fixterm();
+	fixterm();
 #else /* SYSV2 || SYSV3 */
-    cbreak();
-    nonl();
-    noecho ();
+	cbreak();
+	nonl();
+	noecho ();
 #endif /* SYSV2 || SYSV3 */
-    kbd_again();
-    (void) clear();
-    FullUpdate++;
+	kbd_again();
+	if (color && has_colors())
+	    bkgdset(COLOR_PAIR(1) | ' ');
+	(void) clear();
+	FullUpdate++;
+    }
 }
 
 /* clean up curses */
 void
-deraw()
+deraw(int ClearLastLine)
 {
-    (void) move(LINES - 1, 0);
-    (void) clrtoeol();
-    (void) refresh();
+    if (usecurses) {
+	if (ClearLastLine) {
+	    if (color && has_colors())
+		bkgdset(COLOR_PAIR(0) | ' ');
+	    (void) move(lines - 1, 0);
+	    (void) clrtoeol();
+	    (void) refresh();
+	}
 #if SYSV2 || SYSV3
-    resetterm();
+	resetterm();
 #else
-    nocbreak();
-    nl();
-    echo();
+	nocbreak();
+	nl();
+	echo();
 #endif
-    resetkbd();
+	resetkbd();
+    }
 }
 
 #endif /* VMS */
