@@ -1,7 +1,7 @@
 /*	SC	A Spreadsheet Calculator
  *
  *	One line vi emulation
- *	$Revision: 7.7 $
+ *	$Revision: 7.13 $
  */
 
 #include <sys/types.h>
@@ -15,7 +15,6 @@
 
 #include <signal.h>
 #include <curses.h>
-#include <stdio.h>
 #include <ctype.h>
 #include "sc.h"
 
@@ -23,21 +22,22 @@
 
 static void append_line();
 static void back_hist();
-static int  back_line();
-static int  back_word();
+static int  back_line(int arg);
+static int  back_word(int arg);
 static void back_space();
 static void col_0();
 static void cr_line();
-static void delete_cmd(int change);
-static void del_chars(register int first, register int last);
-static void del_in_line(int back_null);
+static void del_in_line(int arg, int back_null);
 static void del_to_end();
+static void dogoto();
+static void dotab();
 static void dotcmd();
-static int  find_char(int n);
+static int  find_char(int arg, int n);
 static void for_hist();
-static int  for_line(int stop_null);
-static int  for_word(int end_word, int stop_null);
+static int  for_line(int arg, int stop_null);
+static int  for_word(int arg, int end_word, int stop_null);
 static void last_col();
+static void change_case(int arg);
 static void rep_char();
 static void replace_in_line(int c);
 static void replace_mode();
@@ -48,12 +48,15 @@ static void search_again();
 static void search_hist();
 static void search_mode();
 static void stop_edit();
-static int  to_char(int n);
+static int  to_char(int arg, int n);
 static void u_save(int c);
+static void yank_cmd(int delete, int change);
+static void yank_chars(register int first, register int last, int delete);
 
-extern int showrange;
+extern int framerows;		/* Rows in current frame */
 extern char mode_ind;		/* Mode indicator */
 extern int lcols;		/* Spreadsheet Column the cursor was in last */
+char *completethis = NULL;
 
 /* values for mode below */
 
@@ -82,27 +85,53 @@ static char	dotb[DOTLEN];
 static int	doti = 0;
 static int	do_dot = 0;
 static char	putbuf[FBUFLEN];
+static int	findfunc = '0';
+static int	findchar = 1;
+static int	finddir = 0;
 
 void
 write_line(int c)
 {
+    struct frange *fr;
+    struct crange *cr;
+
+    if (c != ctl('i')) completethis = NULL;
     if (mode == EDIT_MODE) {
 	switch (c) {
-#ifdef KEY_BACKSPACE
 	case KEY_BACKSPACE:
-#endif
-	case (ctl('h')):	linelim = back_line();			break;
-	case (ctl('i')):	startshow();				break;
+	case (ctl('h')):	linelim = back_line(arg);		break;
+	case (ctl('i')):	dotab();				break;
 	case (ctl('m')):	cr_line();				break;
+	case 'v':
 	case (ctl('v')):	toggle_navigate_mode();			break;
 	case ESC:	stop_edit();					break;
 	case '+':	for_hist();					break;
 	case '-':	back_hist();					break;
-#ifdef KEY_END
 	case KEY_END:
-#endif
+	case (ctl('e')):
 	case '$':	last_col();					break;
 	case '.':	dotcmd();					break;
+	case ';':	if (findfunc)
+			    ungetch(findchar);
+			else
+			    break;
+			findchar = 0;
+			if (findfunc == 'f')
+			    linelim = find_char(arg, finddir);
+			else
+			    linelim = to_char(arg, finddir);
+									break;
+	case ',':	if (findfunc)
+			    ungetch(findchar);
+			else
+			    break;
+			findchar = 0;
+			if (findfunc == 'f')
+			    linelim = find_char(arg, -(finddir));
+			else
+			    linelim = to_char(arg, -(finddir));
+									break;
+	case '~':	u_save(c); change_case(arg);			break;
 #ifdef KEY_FIND
 	case KEY_FIND:
 #endif
@@ -110,105 +139,85 @@ write_line(int c)
 #ifdef KEY_HOME
 	case KEY_HOME:
 #endif
+	case (ctl('a')):
 	case '0':	col_0();					break;
 	case 'A':	u_save(c); last_col(); append_line();		break;
 	case 'C':	u_save(c); del_to_end(); append_line();		break;
 	case 'D':	u_save(c); del_to_end();			break;
-	case 'F':	linelim = find_char(-1);			break;
+	case 'F':	linelim = find_char(arg, -1);			break;
 	case 'G':	if (histp > 0) histp = lasthist; for_hist();	break;
 	case 'I':	u_save(c); col_0(); insert_mode();		break;
 	case 'P':	u_save(c);
 			ins_string(putbuf);
-			linelim = back_line();				break;
-	case 'R':	replace_mode();					break;
-	case 'T':	linelim = to_char(-1);				break;
+			linelim = back_line(1);				break;
+	case 'R':	u_save(c); replace_mode();			break;
+	case 'T':	linelim = to_char(arg, -1);			break;
 	case 'X':	u_save(c); back_space();			break;
+	case 'Y':	yank_chars(linelim, strlen(line), 0);		break;
 	case 'a':	u_save(c); append_line();			break;
-	case 'b':	linelim = back_word();				break;
-	case 'c':	u_save(c); delete_cmd(1); insert_mode();	break;
-	case 'd':	u_save(c); delete_cmd(0);			break;
-	case 'e':	linelim = for_word(1, 0);			break;
-	case 'f':	linelim = find_char(1);				break;
-#ifdef KEY_LEFT
+	case 'b':	linelim = back_word(arg);			break;
+	case 'c':	u_save(c); yank_cmd(1, 1); insert_mode();	break;
+	case 'd':	u_save(c); yank_cmd(1, 0);			break;
+	case 'e':	linelim = for_word(arg, 1, 0);			break;
+	case 'f':	linelim = find_char(arg, 1);			break;
 	case KEY_LEFT:
-#endif
 	case (ctl('b')):
-	case 'h':	linelim = back_line();				break;
-#ifdef KEY_IC
+	case 'h':	linelim = back_line(arg);			break;
 	case KEY_IC:
-#endif
 	case 'i':	u_save(c); insert_mode();			break;
-#ifdef KEY_DOWN
 	case KEY_DOWN:
-#endif
 	case 'j':	for_hist();					break;
-#ifdef KEY_UP
 	case KEY_UP:
-#endif
 	case 'k':	back_hist();					break;
-#ifdef KEY_RIGHT
 	case KEY_RIGHT:
-#endif
 	case (ctl('f')):
 	case ' ':
-	case 'l':	linelim = for_line(0);				break;
+	case 'l':	linelim = for_line(arg, 0);			break;
 	case 'n':	search_again();					break;
 	case 'p':	u_save(c);
-			linelim = for_line(1);
+			linelim = for_line(1, 1);
 			ins_string(putbuf);
-			linelim = back_line();				break;
+			linelim = back_line(1);				break;
 	case 'q':	stop_edit();					break;
 	case 'r':	u_save(c); rep_char();				break;
-	case 's':	u_save(c); del_in_line(0); insert_mode();	break;
-	case 't':	linelim = to_char(1);				break;
+	case 's':	u_save(c); del_in_line(arg, 0); insert_mode();	break;
+	case 't':	linelim = to_char(arg, 1);			break;
 	case 'u':	restore_it();					break;
-	case 'w':	linelim = for_word(0, 0);			break;
-#ifdef KEY_DC
+	case 'w':	linelim = for_word(arg, 0, 0);			break;
 	case KEY_DC:
-#endif
-	case 'x':	u_save(c); del_in_line(1);			break;
+	case 'x':	u_save(c); del_in_line(arg, 1);			break;
+	case 'y':	yank_cmd(0, 0);					break;
 	default:							break;
 	}
     } else if (mode == INSERT_MODE) { 
-	savedot(c);
+	if (c == (ctl('m')))
+	    savedot(ESC);
+	else
+	    savedot(c);
 	switch (c) {
-#ifdef KEY_BACKSPACE
 	case KEY_BACKSPACE:
-#endif
 	case (ctl('h')):	back_space();				break;
-	case (ctl('i')):	startshow();				break;
+	case (ctl('i')):	dotab();				break;
 	case (ctl('m')):	cr_line();				break;
 	case (ctl('v')):	toggle_navigate_mode();			break;
-#ifdef KEY_LEFT
 	case KEY_LEFT:
-#endif
-	case (ctl('b')):	linelim = back_line();			break;
-#ifdef KEY_RIGHT
+	case (ctl('b')):	linelim = back_line(arg);		break;
 	case KEY_RIGHT:
-#endif
-	case (ctl('f')):	linelim = for_line(1);			break;
-#ifdef KEY_DOWN
+	case (ctl('f')):	linelim = for_line(arg, 1);		break;
 	case KEY_DOWN:
-#endif
 	case (ctl('n')):	for_hist();				break;
-#ifdef KEY_UP
 	case KEY_UP:
-#endif
 	case (ctl('p')):	back_hist();				break;
-#ifdef KEY_HOME
-	case KEY_HOME:		col_0();				break;
-#endif
-#ifdef KEY_END
-	case KEY_END:		last_col();				break;
-#endif
+	case KEY_HOME:
+	case (ctl('a')):	col_0();				break;
+	case KEY_END:
+	case (ctl('e')):	last_col();				break;
 	case ESC:		edit_mode();				break;
 	default:		ins_in_line(c);				break;
 	}
     } else if (mode == SEARCH_MODE) {
 	switch (c) {
-#ifdef KEY_BACKSPACE
 	case KEY_BACKSPACE:
-#endif
 	case (ctl('h')):	back_space();				break;
 	case (ctl('m')):	search_hist();				break;
 	case ESC:		edit_mode();				break;
@@ -217,22 +226,32 @@ write_line(int c)
     } else if (mode == REP_MODE) {
 	savedot(c);
 	switch (c) {
-#ifdef KEY_BACKSPACE
 	case KEY_BACKSPACE:
-#endif
-	case (ctl('h')):	back_space();				break;
+	case (ctl('h')):	if (linelim > strlen(undo_line))
+				    back_space();
+				else {
+				    linelim = back_line(1);
+				    *(line+linelim) = *(undo_line+linelim);
+				}					break;
 	case (ctl('m')):	cr_line();				break;
 	case ESC:		edit_mode();				break;
 	default:		replace_in_line(c);			break;
 	}
     } else if (mode == NAVIGATE_MODE) {
 	switch (c) {
-	case ('.'):
+	case '.':
+	case ':':
 	case (ctl('i')):	if (!showrange) {
 				    toggle_navigate_mode();
 				    startshow();
 				} else {
 				    showdr();
+				    ins_in_line(' ');
+				} 					break;
+	case '+':
+	case '-':		if (!showrange) {
+				    ins_string(v_name(currow, curcol));
+				    ins_in_line(c);
 				}					break;
 	case (ctl('m')):	if (!showrange) {
 				    ins_string(v_name(currow, curcol));
@@ -241,74 +260,125 @@ write_line(int c)
 				    toggle_navigate_mode();
 				    cr_line();
 				}					break;
+	case (ctl('a')):	{   /* insert variable value */
+				    struct ent *p = *ATBL(tbl, currow, curcol);
+				    char temp[100];
+
+				    if (p && p->flags & is_valid) {
+					(void) sprintf(temp, "%.*f",
+						precision[curcol], p->v);
+					ins_string(temp);
+				    }
+				}					break;
+	case 'v':
 	case (ctl('v')):	ins_string(v_name(currow, curcol));
 				toggle_navigate_mode();			break;
-#ifdef KEY_LEFT
+	case 'c':		if ((cr = find_crange(currow, curcol))) {
+				    ins_string(r_name(cr->r_left->row,
+					    cr->r_left->col,
+					    cr->r_right->row,
+					    cr->r_right->col));
+				    toggle_navigate_mode();
+				    ins_in_line(' ');
+				    showrange = 0;
+				}					break;
+	case 'f':		if ((fr = find_frange(currow, curcol))) {
+				    ins_string(r_name(fr->or_left->row,
+					    fr->or_left->col,
+					    fr->or_right->row,
+					    fr->or_right->col));
+				    toggle_navigate_mode();
+				    ins_in_line(' ');
+				    showrange = 0;
+				}					break;
+	case 'r':		if ((fr = find_frange(currow, curcol))) {
+				    ins_string(r_name(fr->ir_left->row,
+					    fr->ir_left->col,
+					    fr->ir_right->row,
+					    fr->ir_right->col));
+				    toggle_navigate_mode();
+				    ins_in_line(' ');
+				    showrange = 0;
+				}					break;
 	case KEY_LEFT:
-#endif
-	case (ctl('b')):
 	case 'h':		backcol(arg);				break;
-#ifdef KEY_RIGHT
 	case KEY_RIGHT:
-#endif
-	case (ctl('f')):
 	case ' ':
 	case 'l':		forwcol(arg);				break;
-#ifdef KEY_DOWN
 	case KEY_DOWN:
-#endif
 	case (ctl('n')):
 	case 'j':		forwrow(arg);				break;
-#ifdef KEY_UP
 	case KEY_UP:
-#endif
 	case (ctl('p')):
 	case 'k':		backrow(arg);				break;
 	case 'q':
 	case ctl('g'):
 	case ESC:		toggle_navigate_mode();
 				showrange = 0;				break;
-	case 'H':
-				backcol((curcol-stcol+1)+1);
+	case 'H':		backcol(curcol - stcol + 2);
 									break;
-#ifdef KEY_NPAGE
 	case KEY_NPAGE:			/* next page */
-#endif
+	case (ctl('f')):
 	case 'J':
-				forwrow((LINES-RESROW)/2);
-				strow = strow + (LINES-RESROW)/2;
+				{
+				int ps;
+
+				ps = pagesize ? pagesize :
+				    (LINES - RESROW - framerows)/2;
+				forwrow(arg * ps);
+				strow = strow + (arg * ps);
 				FullUpdate++;
+				}
 									break;
-#ifdef	KEY_PPAGE
 	case KEY_PPAGE:			/* previous page */
-#endif
+	case (ctl('b')):
 	case 'K':
-				backrow((LINES-RESROW)/2);
-				if (strow > (LINES-RESROW)/2)
-				    strow = strow - (LINES-RESROW)/2;
-				else
-				    strow = 0;
+				{
+				int ps;
+
+				ps = pagesize ? pagesize :
+				    (LINES - RESROW - framerows)/2;
+				backrow(arg * ps);
+				strow = strow - (arg * ps);
+				if (strow < 0) strow = 0;
 				FullUpdate++;
+				}
 									break;
 	case 'L':
 				forwcol(lcols -(curcol-stcol)+1);
 									break;
-#ifdef KEY_HOME
 	case KEY_HOME:
+				/* Remember the current position */
+				savedrow[0] = currow;
+				savedcol[0] = curcol;
+				savedstrow[0] = strow;
+				savedstcol[0] = stcol;
+
 				currow = 0;
 				curcol = 0;
 				rowsinrange = 1;
 				colsinrange = fwidth[curcol];
 				FullUpdate++;
 									break;
-#endif
-	case '0':		curcol = 0;
+	case '0':
+				/* Remember the current position */
+				savedrow[0] = currow;
+				savedcol[0] = curcol;
+				savedstrow[0] = strow;
+				savedstcol[0] = stcol;
+
+				curcol = 0;
 		   		rowsinrange = 1;
 		   		colsinrange = fwidth[curcol];		break;
-	case '$':
-				{
+	case '$':		{
 				register struct ent *p;
 
+				/* Remember the current position */
+				savedrow[0] = currow;
+				savedcol[0] = curcol;
+				savedstrow[0] = strow;
+				savedstcol[0] = stcol;
+
 				curcol = maxcols - 1;
 				while (!VALID_CELL(p, currow, curcol) &&
 					curcol > 0)
@@ -318,13 +388,24 @@ write_line(int c)
 				break;
 				}
 	case '^':
+				/* Remember the current position */
+				savedrow[0] = currow;
+				savedcol[0] = curcol;
+				savedstrow[0] = strow;
+				savedstcol[0] = stcol;
+
 				currow = 0;
 				rowsinrange = 1;
 				colsinrange = fwidth[curcol];		break;
-	case '#':
-				{
+	case '#':		{
 				register struct ent *p;
 
+				/* Remember the current position */
+				savedrow[0] = currow;
+				savedcol[0] = curcol;
+				savedstrow[0] = strow;
+				savedstcol[0] = stcol;
+
 				currow = maxrows - 1;
 				while (!VALID_CELL(p, currow, curcol) &&
 					currow > 0)
@@ -332,8 +413,15 @@ write_line(int c)
 				rowsinrange = 1;
 				colsinrange = fwidth[curcol];		break;
 				}
-	case 'w':
-				{
+	case 'm':
+				markcell();				break;
+	case '`': case '\'':
+				dotick(c);				break;
+	case '*':
+				gotonote();				break;
+	case 'g':
+				dogoto();				break;
+	case 'w':		{
 				register struct ent *p;
 
 				while (--arg>=0) {
@@ -358,8 +446,7 @@ write_line(int c)
 				colsinrange = fwidth[curcol];
 				break;
 				}
-	case 'b':
-				{
+	case 'b':		{
 				register struct ent *p;
 
 				while (--arg>=0) {
@@ -394,7 +481,7 @@ edit_mode()
     mode_ind = 'e';
     if (linelim < 0)	/* -1 says stop editing, ...so we still aren't */
 	return;
-    linelim = back_line();
+    linelim = back_line(1);
 }
 
 void
@@ -429,15 +516,21 @@ toggle_navigate_mode()
 
     switch (prev_mode) {
 	case INSERT_MODE:
-				prev_mode = NAVIGATE_MODE;
-				insert_mode();
-				break;
+				if (mode == NAVIGATE_MODE) {
+				    prev_mode = NAVIGATE_MODE;
+				    insert_mode();
+				    break;
+				} else
+				    return;
 	case EDIT_MODE:
-				prev_mode = NAVIGATE_MODE;
-				limtmp = linelim;
-				edit_mode();
-				linelim = limtmp;
-				break;
+				if (mode == NAVIGATE_MODE) {
+				    prev_mode = NAVIGATE_MODE;
+				    limtmp = linelim;
+				    edit_mode();
+				    linelim = limtmp;
+				    break;
+				} else
+				    return;
 	case NAVIGATE_MODE:
 				prev_mode = mode;
 				mode_ind = 'v';
@@ -449,6 +542,46 @@ toggle_navigate_mode()
     }
 }
 
+void
+dotab()
+{
+    static struct range *firstmatch;
+    static struct range *lastmatch;
+    static struct range *nextmatch;
+    int len;
+
+    if (linelim > 1 && isalnum(line[linelim-1]) || line[linelim-1] == '_' ||
+	    (completethis && line[linelim-1] == ' ')) {
+	if (!completethis) {
+	    for (completethis = line + linelim - 1; isalnum(*completethis) ||
+		    *completethis == '_'; completethis--) /* */;
+	    completethis++;
+	    len = line + linelim - completethis;
+	    if (!find_range(completethis, len, NULL, NULL, &lastmatch)) {
+		firstmatch = lastmatch;
+		while (firstmatch->r_next &&
+			!strncmp(completethis, firstmatch->r_next->r_name, len))
+		    firstmatch = firstmatch->r_next;
+		nextmatch = firstmatch;
+	    } else
+		nextmatch = NULL;
+	}
+	if (nextmatch) {
+	    len = line + linelim - completethis;
+	    memmove(completethis, line + linelim, strlen(line) - linelim + 1);
+	    linelim -= len;
+	    ins_string(nextmatch->r_name);
+	    if (*(completethis - 1) == ' ' && *(line + linelim) != ' ')
+		ins_in_line(' ');
+	    if (nextmatch == lastmatch)
+		nextmatch = firstmatch;
+	    else
+		nextmatch = nextmatch->r_prev;
+	}
+    } else
+	startshow();
+}
+
 /* show the current range (see ^I), we are moving around to define a range */
 void
 startshow()
@@ -460,18 +593,48 @@ startshow()
 }
 
 /* insert the range we defined by moving around the screen, see startshow() */
+
+#define SHOWROWS 2
+#define SHOWCOLS 3
+
 void
 showdr()
 {
-    int		minsr, minsc, maxsr, maxsc;
-    char	*p;
+    int			minsr, minsc, maxsr, maxsc;
+    char		*p;
+    char		r[12];
+    struct frange	*fr = find_frange(currow, curcol);
 
-    minsr = showsr < currow ? showsr : currow;
-    minsc = showsc < curcol ? showsc : curcol;
-    maxsr = showsr > currow ? showsr : currow;
-    maxsc = showsc > curcol ? showsc : curcol;
-    toggle_navigate_mode();
-    ins_string(r_name(minsr, minsc, maxsr, maxsc));
+    if (showrange == SHOWROWS) {
+	minsr = showsr < currow ? showsr : currow;
+	minsc = fr ? fr->or_left->col : 0;
+	maxsr = showsr > currow ? showsr : currow;
+	maxsc = fr ? fr->or_right->col : maxcols;
+
+	toggle_navigate_mode();
+	sprintf(r, "%d:%d", minsr, maxsr);
+	ins_string(r);
+    } else if (showrange == SHOWCOLS) {
+	minsr = 0;
+	minsc = showsc < curcol ? showsc : curcol;
+	maxsr = maxrows;
+	maxsc = showsc > curcol ? showsc : curcol;
+
+	strcpy(r, coltoa(minsc));
+	strcat(r, ":");
+	strcat(r, coltoa(maxsc));
+
+	toggle_navigate_mode();
+	ins_string(r);
+    } else {
+	minsr = showsr < currow ? showsr : currow;
+	minsc = showsc < curcol ? showsc : curcol;
+	maxsr = showsr > currow ? showsr : currow;
+	maxsc = showsc > curcol ? showsc : curcol;
+
+	toggle_navigate_mode();
+	ins_string(r_name(minsr, minsc, maxsr, maxsc));
+    }
     showrange = 0;
 }
 
@@ -533,7 +696,6 @@ vigetch()
 	}
     }
     c = nmgetch();
-    savedot(c);
     return (c);
 }
 
@@ -602,13 +764,21 @@ stop_edit()
  * the last character of the line.
  */
 static int
-for_line(int stop_null)
+for_line(int arg, int stop_null)
 {
-    if (linelim >= 0 && line[linelim] != '\0' && 
-	    (line[linelim+1] != '\0' || stop_null))
-	return (linelim+1);
-    else
+    int cpos = linelim;
+
+    if (linelim < 0)
 	return (linelim);
+    else if (linelim + arg <= strlen(line))
+	cpos += arg;
+    else
+	cpos = strlen(line);
+
+    if (cpos == strlen(line) && cpos > 0 && !stop_null)
+	return (cpos - 1);
+    else
+	return (cpos);
 }
 
 /* If end_word is non-zero, go to next end-of-word.  Otherwise, go to next
@@ -616,86 +786,90 @@ for_line(int stop_null)
  */
 
 static int
-for_word(int end_word, int stop_null)
+for_word(int arg, int end_word, int stop_null)
 {
     register int c;
     register int cpos;
 
     cpos = linelim;
 
-    if (end_word)
-	cpos++;
-
-    if (line[cpos] == ' ') {
-	while (line[cpos] == ' ')
+    while (cpos < strlen(line) && arg--) {
+	if (end_word)
 	    cpos++;
-	if (cpos > 0 && line[cpos] == '\0')
+
+	if (line[cpos] == ' ') {
+	    while (line[cpos] == ' ')
+		cpos++;
+	    if (cpos > 0 && line[cpos] == '\0')
+		--cpos;
+	    if (!end_word)
+		continue;
+	}
+
+	if (istext(line[cpos])) {
+	    while ((c = line[cpos]) && istext(c)) 
+		cpos++;
+	} else {
+	    while ((c = line[cpos]) && !istext(c) && c != ' ')
+		cpos++;
+	}
+
+	if (end_word)
+	    cpos--;
+	else while (line[cpos] == ' ')
+	    cpos++;
+
+	if (cpos > 0 && line[cpos] == '\0' && !stop_null) 
 	    --cpos;
-	if (!end_word)
-	    return(cpos);
     }
 
-    if (istext(line[cpos])) {
-    	while ((c = line[cpos]) && istext(c)) 
-	    cpos++;
-    } else {
-	while ((c = line[cpos]) && !istext(c) && c != ' ')
-	    cpos++;
-    }
-
-    if (end_word)
-	cpos--;
-    else while (line[cpos] == ' ')
-        cpos++;
-
-    if (cpos > 0 && line[cpos] == '\0' && !stop_null) 
-        --cpos;
-
-    return(cpos);
+    return (cpos);
 }
 
 static int
-back_line()
+back_line(int arg)
 {
-    if (linelim)
-        return(linelim-1);
+    if (linelim > arg)
+        return (linelim - arg);
     else
-	return(0);
+	return (0);
 }
 
 static int
-back_word()
+back_word(int arg)
 {
     register int c;
     register int cpos;
 
     cpos = linelim;
 
-    if (line[cpos] == ' ') {
-	/* Skip white space */
-        while (cpos > 0 && line[cpos] == ' ')
-	    --cpos;
-    } else if (cpos > 0 && (line[cpos-1] == ' ' 
-		     || ( istext(line[cpos]) && !istext(line[cpos-1]))
-		     || (!istext(line[cpos]) &&  istext(line[cpos-1])))) {
-	/* Started on the first char of a word - back up to prev. word */
-	--cpos;
-        while (cpos > 0 && line[cpos] == ' ')
-	    --cpos;
-    }
-
-    /* Skip across the word - goes 1 too far */
-    if (istext(line[cpos])) {
-    	while (cpos > 0 && (c = line[cpos]) && istext(c)) 
+    while (cpos > 0 && arg--) {
+	if (line[cpos] == ' ') {
+	    /* Skip white space */
+	    while (cpos > 0 && line[cpos] == ' ')
 		--cpos;
-    } else {
-	while (cpos > 0 && (c = line[cpos]) && !istext(c) && c != ' ')
+	} else if (cpos > 0 && (line[cpos-1] == ' ' 
+			|| ( istext(line[cpos]) && !istext(line[cpos-1]))
+			|| (!istext(line[cpos]) &&  istext(line[cpos-1])))) {
+	    /* Started on the first char of a word - back up to prev. word */
+	    --cpos;
+	    while (cpos > 0 && line[cpos] == ' ')
 		--cpos;
-    }
+	}
 
-    /* We are done - fix up the one too far */
-    if (cpos > 0 && line[cpos] && line[cpos+1]) 
-	cpos++;
+	/* Skip across the word - goes 1 too far */
+	if (istext(line[cpos])) {
+	    while (cpos > 0 && (c = line[cpos]) && istext(c)) 
+		    --cpos;
+	} else {
+	    while (cpos > 0 && (c = line[cpos]) && !istext(c) && c != ' ')
+		    --cpos;
+	}
+
+	/* We are done - fix up the one too far */
+	if (cpos > 0 && line[cpos] && line[cpos+1]) 
+	    cpos++;
+    }
 
     return(cpos);
 }
@@ -706,18 +880,20 @@ back_word()
  * line terminator.  Otherwise, don't.
  */
 static void
-del_in_line(int back_null)
+del_in_line(int arg, int back_null)
 {
     register int len, i;
 
     if (linelim >= 0) {
 	len = strlen(line);
+	if (arg > len - linelim)
+	    arg = len - linelim;
 	if (linelim == len && linelim > 0)
 	    linelim--;
-	putbuf[0] = line[linelim];
-	putbuf[1] = '\0';
+	strncpy(putbuf, line + linelim, arg);
+	putbuf[arg] = '\0';
 	for (i = linelim; i < len; i++)
-	    line[i] = line[i+1];
+	    line[i] = line[i+arg];
     }
     if (back_null && linelim > 0 && line[linelim] == '\0')
 	--linelim;
@@ -759,6 +935,22 @@ append_line()
     insert_mode();
 }
 
+static void
+change_case(arg)
+{
+    if (linelim < 0) {
+    	linelim = 0;
+	*line = '\0';
+    }
+    while (arg--) {
+	if (islower(line[linelim]))
+	    line[linelim] = toupper(line[linelim]);
+	else if (isupper(line[linelim]))
+	    line[linelim] = tolower(line[linelim]);
+	linelim = for_line(1, 0);
+    }
+}
+
 static void
 rep_char()
 {
@@ -769,6 +961,7 @@ rep_char()
 	*line = '\0';
     }
     c = vigetch();
+    savedot(c);
     if (c < 256 && c!= ESC && c != ctl('g')) {
 	if (line[linelim] == '\0')
 	    line[linelim+1] = '\0';
@@ -800,12 +993,12 @@ back_space()
 	return;
 
     if (line[linelim] == '\0') {
-	linelim = back_line();
-	del_in_line(1);
+	linelim = back_line(1);
+	del_in_line(1, 1);
 	linelim = strlen(line);
     } else {
-	linelim = back_line();
-	del_in_line(1);
+	linelim = back_line(1);
+	del_in_line(1, 1);
     }
 }
 
@@ -817,28 +1010,38 @@ int
 get_motion(int change)
 {
     int c;
+    int arg2 = 0;
 
     c = vigetch();
+    savedot(c);
+    if (c == 0)		return (0);
+    while (c >= '0' && c <= '9') {
+	arg2 = 10 * arg2 + c - '0';
+	c = vigetch();
+	savedot(c);
+    }
+    if (!arg2)
+	arg2++;
+    arg *= arg2;
     switch (c) {
-	case '0':	return (0);
 	case '$':	return (strlen(line));
-	case 'b':	return (back_word());
+	case 'b':	return (back_word(arg));
 	case 'c':	return (change ? -1 : linelim);
 	case 'd':	return (!change ? -1 : linelim);
-	case 'e':	return (for_word(1, 1) + 1);
-	case 'f':	return ((c = find_char(1)) == linelim ? c : c + 1);
-	case 'F':	return (find_char(-1));
-	case 'h':	return (back_line());
-	case 'l':	return (for_line(1));
-	case 't':	return ((c = to_char(1)) == linelim ? c : c + 1);
-	case 'T':	return (to_char(-1));
-	case 'w':	return (for_word(change, 1) + change);
+	case 'e':	return (for_word(arg, 1, 1) + 1);
+	case 'f':	return ((c = find_char(arg, 1)) == linelim ? c : c + 1);
+	case 'F':	return (find_char(arg, -1));
+	case 'h':	return (back_line(arg));
+	case 'l':	return (for_line(arg, 1));
+	case 't':	return ((c = to_char(arg, 1)) == linelim ? c : c + 1);
+	case 'T':	return (to_char(arg, -1));
+	case 'w':	return (for_word(arg, change, 1) + change);
 	default:	return (linelim);
     }
 }
 
 static void
-delete_cmd(int change)
+yank_cmd(int delete, int change)
 {
     int cpos;
 
@@ -846,11 +1049,11 @@ delete_cmd(int change)
 	cpos++;
 	linelim = strlen(line);
     }
-    del_chars(cpos, linelim);
+    yank_chars(cpos, linelim, delete);
 }
 
 static void
-del_chars(register int first, register int last)
+yank_chars(register int first, register int last, int delete)
 {
     int temp;
 
@@ -864,7 +1067,8 @@ del_chars(register int first, register int last)
     linelim = first;
     strncpy(putbuf, line + first, last - first);
     putbuf[last - first] = '\0';
-    memmove(line + first, line + last, strlen(line + last) + 1);
+    if (delete)
+	memmove(line + first, line + last, strlen(line + last) + 1);
 }
 
 static void
@@ -874,12 +1078,14 @@ del_to_end()
 	return;
     strcpy(putbuf, line + linelim);
     line[linelim] = '\0';
-    linelim = back_line();
+    linelim = back_line(1);
 }
 
 static void
 cr_line()
 {
+    struct frange *fr;
+
     insert_mode();
     if (linelim != -1) {
 	showrange = 0;
@@ -895,7 +1101,14 @@ cr_line()
 			forwcol(1);
 			currow = 0;
 		    } else {
-			forwrow(1);
+			if (autoinsert && (fr = find_frange(currow, curcol))) {
+			    forwrow(1);
+			    if (currow > fr->ir_right->row) {
+				backrow(1);
+				insertrow(1, 1);
+			    }
+			} else
+			    forwrow(1);
 		    }
 		    break;
 		case CRCOLS:
@@ -903,7 +1116,14 @@ cr_line()
 			forwrow(1);
 			curcol = 0;
 		    } else {
-			forwcol(1);
+			if (autoinsert && (fr = find_frange(currow, curcol))) {
+			    forwcol(1);
+			    if (curcol > fr->ir_right->col) {
+				backcol(1);
+				insertcol(1, 1);
+			    }
+			} else
+			    forwcol(1);
 		    }
 		    break;
 		default:
@@ -965,7 +1185,7 @@ back_hist()
 
     if (lasthist >= 0) {
     	(void) strcpy(line, history[histp].histline);
-	linelim = 0;
+	last_col();
     }
 }
 
@@ -1034,7 +1254,7 @@ for_hist()
 
     if (lasthist >= 0) {
 	(void) strcpy(line, history[histp].histline);
-	linelim = 0;
+	last_col();
     }
 }
 
@@ -1053,29 +1273,127 @@ last_col()
 }
 
 static int
-find_char(int n)
+find_char(int arg, int n)
 {
-    register int c;
     register int i;
 
-
-    c = vigetch();
+    if (findchar)
+	finddir = n;
+    findchar = vigetch();
+    switch (dotb[doti - 1]) {
+	case 'f': case 'F': case 't': case 'T':
+	    savedot(findchar);
+	default:
+	    break;
+    }
     i = linelim;
-    while (i >= 0 && line[i] && line[i] != c)
+    while (arg--) {
 	i += n;
-    if (i < 0 || !line[i])
-	i = linelim;
+	while (i >= 0 && line[i] && line[i] != findchar)
+	    i += n;
+	if (i < 0 || !line[i]) {
+	    i = linelim;
+	    break;
+	}
+    }
+    findfunc = 'f';
     return (i);
 }
 
 static int
-to_char(int n)
+to_char(int arg, int n)
 {
     register int i;
+    int tmp = linelim;
 
-    i = find_char(n);
+    if (linelim + n >= 0 && linelim + n < strlen(line))
+	linelim += n;
+    i = find_char(arg, n);
     if (i != linelim)
 	i -= n;
+    linelim = tmp;
+    findfunc = 't';
 
     return (i);
 }
+
+static void
+dogoto()
+{
+    static char		*tempc = NULL;
+    static unsigned	templen = 0;
+    int			tempi;
+
+    if (strlen(line) + 1 > templen) {
+    	templen = strlen(line) + 40;
+	tempc = scxrealloc(tempc, templen);
+    }
+
+    strcpy(tempc, line);
+    tempi = linelim;
+
+    /* Can't switch back to navigate mode if insert_mode() is used here
+     * instead of toggle_navigate_mode(), which is what we want when doing
+     * a goto from within navigate mode.
+     */
+    insert_mode();
+    /* Tempted as I was, I had to resist making this "Where would you like
+     * to go today?" - CRM :)
+     */
+    query("goto where?", NULL);
+    if (linelim >= 0) {
+	memmove(line + 5, line, strlen(line) + 1);
+	strncpy(line, "goto ", 5);
+	linelim = 0;
+	yyparse();
+    }
+
+    strcpy(line, tempc);
+    linelim = tempi;
+    /* Now we need to change back to navigate mode ourselves so that
+     * toggle_navigate_mode() will work properly again.
+     */
+    mode_ind = 'v';
+    mode = NAVIGATE_MODE;
+    if (!showrange)
+	toggle_navigate_mode();
+}
+
+void
+query(char *s, char *data)
+{
+    int c;
+
+    insert_mode();
+    if (data != NULL) {
+	strcpy(line, data);
+	linelim = strlen(line);
+    } else {
+    	*line = '\0';
+	linelim = 0;
+    }
+    if (s != NULL) error(s);
+
+    while (linelim >= 0) {
+	update(0);
+	switch (c = nmgetch()) {
+	    case ctl('m'):
+		error("");
+		return;
+	    case ctl('g'):
+		line[0] = '\0';
+		linelim = -1;
+		error("");
+		update(0);
+		return;
+	    case ctl('l'):
+		FullUpdate++;
+		clearok(stdscr,1);
+		update(1);
+		if (s != NULL) error(s);
+		break;
+	    default:
+		write_line(c);
+	}
+    }
+}
